-- db/init_db.sql.  Generated from init_db.sql.in by configure.
-- -----------------------------------------------------------------------------------------
-- This file initializes the psyche database
-- It must be run as the database superuser (postgres)
-- -----------------------------------------------------------------------------------------

\c postgres

CREATE ROLE psyche_col;
CREATE ROLE psyche_ui;
CREATE ROLE pfcapd LOGIN ;
CREATE ROLE psyche LOGIN ;
ALTER GROUP psyche_col ADD USER pfcapd;   --add collector user to psyche_col
ALTER GROUP psyche_ui ADD USER psyche;    --add GUI user to psyche_ui

CREATE DATABASE psyche;
GRANT connect ON DATABASE psyche TO psyche_col, psyche_ui;

\c psyche

-- -----------------------------------------------------------------------------------------
-- TABLE SCHEMAS
-- -----------------------------------------------------------------------------------------

-- --------------------
-- flow table
-- holds raw data imported directly from collector
-- --------------------
CREATE TABLE flow 
(
    id bigserial primary key,  -- unique id for each flow
    router character varying(16) not null,  -- IP address of router/exporter that reported this flow
    src_addr inet not null,   -- source IP address of flow
    dst_addr inet not null,   -- destination IP address of flow
    start_time timestamp with time zone not null,  -- flow's start time
    end_time timestamp with time zone not null,   -- flow's completion time
    tcp_flags integer default 0,  -- cumulative tcp flags for flow
    prot integer not null,  -- protocol number
    tos integer default 0,  -- IP type of service
    input_if integer default 0,  -- input SNMP interface
    output_if integer default 0,  -- output SNMP interface
    src_as integer default 0,  -- source autonomous system id 
    dst_as integer default 0,  -- destination autonomous system id
    src_port integer default 0,  -- source port number of flow
    dst_port integer default 0,  -- destination port number of flow
    src_mask integer default 0,  -- source address prefix mask bits
    dst_mask integer default 0,  -- destination address prefix mask bits
    next_hop inet default '0.0.0.0',  -- next hop router's IP address
    pkts bigint not null,  -- number of packets in flow
    bytes bigint not null  -- number of bytes in flow
);
CREATE INDEX flow_start_time_idx ON flow("start_time"); 

-- --------------------
-- collector state table
-- holds configuration values used by the collector-invoked stored procedures
-- ---------------------- 
CREATE TABLE col_state 
(
    id integer primary key check(id = 1), -- unique id; only one row allowed in table
    interval_length interval default(interval '300') check(interval_length = interval '300'), -- interval used for aggregation purposes
    tcp_age_interval interval default '30 minutes' 
	check(tcp_age_interval BETWEEN '1 minute'::interval AND '61 minutes'::interval), -- minutes of inactivity before a session is aged
    udp_age_interval interval default '30 minutes' 
	check(udp_age_interval BETWEEN '1 minute'::interval AND '61 minutes'::interval), -- minutes of inactivity before a session is aged
    other_age_interval interval default '30 minutes'
	check(other_age_interval BETWEEN '1 minute'::interval AND '61 minutes'::interval) -- minutes of inactivity before a session is aged
);
INSERT INTO col_state(id) VALUES(1);  -- initialize default configuration values

-- --------------------
-- source traffic aggregation state table
-- holds state info used by the aggregate_src_traffic stored procedure
-- ---------------------- 
CREATE TABLE src_traffic_state 
(
    id integer primary key check(id = 1), -- unique id; only one row allowed in table
    last_fid bigint default 0   -- last-processed flow table id for src_traffic aggregation
);
INSERT INTO src_traffic_state(id) VALUES(1);  -- initialize state values

-- --------------------
-- destination traffic aggregation state table
-- holds state info used by the aggregate_dst_traffic stored procedure
-- ---------------------- 
CREATE TABLE dst_traffic_state 
(
    id integer primary key check(id = 1), -- unique id; only one row allowed in table
    last_fid bigint default 0   -- last-processed flow table id for dst_traffic aggregation
);
INSERT INTO dst_traffic_state(id) VALUES(1);  -- initialize state values

-- --------------------
-- tcp session aggregation state table
-- holds state info used by the aggregate_tcp_session stored procedure
-- ---------------------- 
CREATE TABLE tcp_session_state 
(
    id integer primary key check(id = 1), -- unique id; only one row allowed in table
    last_fid bigint default 0    -- last-processed flow table id for tcp_session aggregation
);
INSERT INTO tcp_session_state(id) VALUES(1);  -- initialize state values

-- --------------------
-- udp session aggregation state table
-- holds state info used by the aggregate_udp_session stored procedure
-- ---------------------- 
CREATE TABLE udp_session_state 
(
    id integer primary key check(id = 1), -- unique id; only one row allowed in table
    last_fid bigint default 0    -- last-processed flow table id for udp_session aggregation
);
INSERT INTO udp_session_state(id) VALUES(1);  -- initialize state values

-- --------------------
-- other session aggregation state table
-- holds state info used by the aggregate_other_session stored procedure
-- ---------------------- 
CREATE TABLE other_session_state 
(
    id integer primary key check(id = 1), -- unique id; only one row allowed in table
    last_fid bigint default 0    -- last-processed flow table id for other_session aggregation
);
INSERT INTO other_session_state(id) VALUES(1);  -- initialize state values

-- --------------------
-- tcp session aging state table
-- holds state info used by the age_tcp_session stored procedure
-- ---------------------- 
CREATE TABLE aged_tcp_session_state 
(
    id integer primary key check(id = 1) -- unique id; only one row allowed in table
);
INSERT INTO aged_tcp_session_state(id) VALUES(1);  -- initialize state values

-- --------------------
-- udp session aging state table
-- holds state info used by the age_udp_session stored procedure
-- ---------------------- 
CREATE TABLE aged_udp_session_state 
(
    id integer primary key check(id = 1) -- unique id; only one row allowed in table
);
INSERT INTO aged_udp_session_state(id) VALUES(1);  -- initialize state values

-- --------------------
-- other session aging state table
-- holds state info used by the age_other_session stored procedure
-- ---------------------- 
CREATE TABLE aged_other_session_state 
(
    id integer primary key check(id = 1) -- unique id; only one row allowed in table
);
INSERT INTO aged_other_session_state(id) VALUES(1);  -- initialize state values

-- --------------------
-- source traffic table
-- holds aggregate source traffic data (by interval) for each unique IP/port/prot triple seen
-- --------------------
CREATE TABLE src_traffic 
(
    id bigserial primary key, -- unique id for each row
    interval_start timestamp with time zone, -- start time of interval
    half_hour_start timestamp with time zone, -- start time of half hour
    hour_start timestamp with time zone, -- start time of hour
    day_start timestamp with time zone, -- start time of day
    week_start timestamp with time zone, -- start time of week
    router character varying(16) not null,  -- IP address of router/exporter
    addr inet not null, -- IP address
    port integer default 0,  --port number
    prot integer default 0,  --protocol number
    bytes bigint default 0, -- number of bytes entering this addr/port over this interval
    pkts bigint default 0, -- number of packets entering this addr/port over this interval
    last_update timestamp with time zone default now(), -- last time this row was updated
    unique(interval_start, router, addr, port, prot) with (fillfactor = 80)  -- aggregate unique key for each row 
) with (fillfactor = 80);
CREATE INDEX src_traffic_interval_idx ON src_traffic("interval_start"); 
CREATE INDEX src_traffic_half_hour_idx ON src_traffic("half_hour_start"); 
CREATE INDEX src_traffic_hour_idx ON src_traffic("hour_start"); 

-- --------------------
-- destination traffic table
-- holds aggregate destination traffic data (by interval) for each unique IP/port/prot triple seen
-- --------------------
CREATE TABLE dst_traffic 
(
    id bigserial primary key, -- unique id for each row
    interval_start timestamp with time zone, -- start time of interval
    half_hour_start timestamp with time zone, -- start time of half hour
    hour_start timestamp with time zone, -- start time of hour
    day_start timestamp with time zone, -- start time of day
    week_start timestamp with time zone, -- start time of week
    router character varying(16) not null,  -- IP address of router/exporter
    addr inet not null, -- IP address
    port integer default 0,  --port number
    prot integer default 0,  --protocol number
    bytes bigint default 0, -- number of bytes leaving this addr/port over this interval
    pkts bigint default 0, -- number of packets leaving this addr/port over this interval
    last_update timestamp with time zone default now(), -- last time this row was updated
    unique(interval_start, router, addr, port, prot) with (fillfactor = 80)  -- aggregate unique key for each row 
) with (fillfactor = 80);
CREATE INDEX dst_traffic_interval_idx ON dst_traffic("interval_start"); 
CREATE INDEX dst_traffic_half_hour_idx ON dst_traffic("half_hour_start"); 
CREATE INDEX dst_traffic_hour_idx ON dst_traffic("hour_start"); 

-- --------------------
-- tcp session table
-- holds reconstructed tcp sessions
-- --------------------
CREATE TABLE tcp_session 
(
    id bigserial primary key,  -- unique id for each row
    session_start timestamp with time zone,  -- start time of tcp session
    interval_start timestamp with time zone, -- start time of interval
    hour_start timestamp with time zone, -- start time of hour
    day_start timestamp with time zone, -- start time of day
    week_start timestamp with time zone, -- start time of week
    last_activity timestamp with time zone,  -- last activity seen for this tcp session
    last_interval timestamp with time zone, -- start time of last interval of activity
    last_hour timestamp with time zone, -- start time of last hour of activity
    last_day timestamp with time zone, -- start time of last day of activity
    last_week timestamp with time zone, -- start time of last week of activity
    router character varying(16),  -- IP address of router/exporter
    src_addr inet not null,  -- source IP address
    dst_addr inet not null,  -- destination IP address
    src_port integer default 0,  -- source port number
    dst_port integer default 0,  -- destination port number
    in_pkts bigint default 0,  -- number of packets entering the src_addr/src_port
    out_pkts bigint default 0,  -- number of packets leaving the src_addr/src_port
    in_bytes bigint default 0,  -- number of bytes entering the src_addr/src_port 
    out_bytes bigint default 0,  -- number of bytes leaving the src_addr/src_port
    last_update timestamp with time zone default now(), -- last time this row was updated
    unique(session_start, router, src_addr, dst_addr, src_port, dst_port) with (fillfactor = 90) -- aggregate unique key for each row
) with (fillfactor = 90);
CREATE INDEX tcp_session_addr_port_idx ON tcp_session("router", "src_addr", "dst_addr", "src_port", "dst_port");
CREATE INDEX tcp_session_interval_idx ON tcp_session("interval_start");
CREATE INDEX tcp_session_last_activity_idx ON tcp_session("last_activity");

-- --------------------
-- aged tcp session table
-- holds aged tcp sessions (see age_tcp_session() to see aging process)
-- --------------------
CREATE TABLE aged_tcp_session 
(
    id bigserial primary key,  -- unique id for each row
    session_start timestamp with time zone,  -- start time of tcp session
    interval_start timestamp with time zone, -- start time of interval
    hour_start timestamp with time zone, -- start time of hour
    day_start timestamp with time zone, -- start time of day
    week_start timestamp with time zone, -- start time of week
    last_activity timestamp with time zone,  -- last activity seen for this tcp session
    last_interval timestamp with time zone, -- start time of last interval of activity
    last_hour timestamp with time zone, -- start time of last hour of activity
    last_day timestamp with time zone, -- start time of last day of activity
    last_week timestamp with time zone, -- start time of last week of activity
    router character varying(16),  -- IP address of router/exporter
    src_addr inet not null,  -- source IP address
    dst_addr inet not null,  -- destination IP address
    src_port integer default 0,  -- source port number
    dst_port integer default 0,  -- destination port number
    in_pkts bigint default 0,  -- number of packets entering the src_addr/src_port
    out_pkts bigint default 0,  -- number of packets leaving the src_addr/src_port
    in_bytes bigint default 0,  -- number of bytes entering the src_addr/src_port 
    out_bytes bigint default 0,  -- number of bytes leaving the src_addr/src_port
    last_update timestamp with time zone default now() -- last time this row was updated
) with (fillfactor = 90);
CREATE INDEX aged_tcp_session_addr_port_idx ON aged_tcp_session("router", "src_addr", "dst_addr", "src_port", "dst_port");
CREATE INDEX aged_tcp_session_interval_idx ON aged_tcp_session("interval_start");
CREATE INDEX aged_tcp_session_last_activity_idx ON aged_tcp_session("last_activity");

-- --------------------
-- udp session table
-- holds reconstructed udp sessions
-- --------------------
CREATE TABLE udp_session 
(
    id bigserial primary key,  -- unique id for each row
    session_start timestamp with time zone,  -- start time of tcp session
    interval_start timestamp with time zone, -- start time of interval
    hour_start timestamp with time zone, -- start time of hour
    day_start timestamp with time zone, -- start time of day
    week_start timestamp with time zone, -- start time of week
    last_activity timestamp with time zone,  -- last activity seen for this tcp session
    last_interval timestamp with time zone, -- start time of last interval of activity
    last_hour timestamp with time zone, -- start time of last hour of activity
    last_day timestamp with time zone, -- start time of last day of activity
    last_week timestamp with time zone, -- start time of last week of activity
    router character varying(16),  -- IP address of router/exporter
    src_addr inet not null,  -- source IP address
    dst_addr inet not null,  -- destination IP address
    src_port integer default 0,  -- source port number
    dst_port integer default 0,  -- destination port number
    in_pkts bigint default 0,  -- number of packets entering the src_addr/src_port
    out_pkts bigint default 0,  -- number of packets leaving the src_addr/src_port
    in_bytes bigint default 0,  -- number of bytes entering the src_addr/src_port 
    out_bytes bigint default 0,  -- number of bytes leaving the src_addr/src_port
    last_update timestamp with time zone default now(), -- last time this row was updated
    unique(session_start, router, src_addr, dst_addr, src_port, dst_port) with (fillfactor = 90) -- aggregate unique key for each row
) with (fillfactor = 90);
CREATE INDEX udp_session_addr_port_idx ON udp_session("router", "src_addr", "dst_addr", "src_port", "dst_port");
CREATE INDEX udp_session_interval_idx ON udp_session("interval_start");
CREATE INDEX udp_session_last_activity_idx ON udp_session("last_activity");

-- --------------------
-- aged udp session table
-- holds aged udp sessions (see age_udp_session() to see aging process)
-- --------------------
CREATE TABLE aged_udp_session 
(
    id bigserial primary key,  -- unique id for each row
    session_start timestamp with time zone,  -- start time of tcp session
    interval_start timestamp with time zone, -- start time of interval
    hour_start timestamp with time zone, -- start time of hour
    day_start timestamp with time zone, -- start time of day
    week_start timestamp with time zone, -- start time of week
    last_activity timestamp with time zone,  -- last activity seen for this tcp session
    last_interval timestamp with time zone, -- start time of last interval of activity
    last_hour timestamp with time zone, -- start time of last hour of activity
    last_day timestamp with time zone, -- start time of last day of activity
    last_week timestamp with time zone, -- start time of last week of activity
    router character varying(16),  -- IP address of router/exporter
    src_addr inet not null,  -- source IP address
    dst_addr inet not null,  -- destination IP address
    src_port integer default 0,  -- source port number
    dst_port integer default 0,  -- destination port number
    in_pkts bigint default 0,  -- number of packets entering the src_addr/src_port
    out_pkts bigint default 0,  -- number of packets leaving the src_addr/src_port
    in_bytes bigint default 0,  -- number of bytes entering the src_addr/src_port 
    out_bytes bigint default 0,  -- number of bytes leaving the src_addr/src_port
    last_update timestamp with time zone default now() -- last time this row was updated
) with (fillfactor = 90);
CREATE INDEX aged_udp_session_addr_port_idx ON aged_udp_session("router", "src_addr", "dst_addr", "src_port", "dst_port");
CREATE INDEX aged_udp_session_interval_idx ON aged_udp_session("interval_start");
CREATE INDEX aged_udp_session_last_activity_idx ON aged_udp_session("last_activity");

-- --------------------
-- other session table
-- holds reconstructed non-tcp and non-udp sessions
-- --------------------
CREATE TABLE other_session 
(
    id bigserial primary key,  -- unique id for each row
    session_start timestamp with time zone,  -- start time of tcp session
    interval_start timestamp with time zone, -- start time of interval
    hour_start timestamp with time zone, -- start time of hour
    day_start timestamp with time zone, -- start time of day
    week_start timestamp with time zone, -- start time of week
    last_activity timestamp with time zone,  -- last activity seen for this tcp session
    last_interval timestamp with time zone, -- start time of last interval of activity
    last_hour timestamp with time zone, -- start time of last hour of activity
    last_day timestamp with time zone, -- start time of last day of activity
    last_week timestamp with time zone, -- start time of last week of activity
    router character varying(16),  -- IP address of router/exporter
    src_addr inet not null,  -- source IP address
    dst_addr inet not null,  -- destination IP address
    src_port integer default 0,  -- source port number
    dst_port integer default 0,  -- destination port number
    prot integer not null,  -- protocol number
    in_pkts bigint default 0,  -- number of packets entering the src_addr/src_port
    out_pkts bigint default 0,  -- number of packets leaving the src_addr/src_port
    in_bytes bigint default 0,  -- number of bytes entering the src_addr/src_port 
    out_bytes bigint default 0,  -- number of bytes leaving the src_addr/src_port
    last_update timestamp with time zone default now(), -- last time this row was updated
    unique(session_start, router, src_addr, dst_addr, src_port, dst_port, prot) with (fillfactor = 90) -- aggregate unique key for each row
) with (fillfactor = 90);
CREATE INDEX other_session_addr_port_idx ON other_session("router", "src_addr", "dst_addr", "src_port", "dst_port", "prot");
CREATE INDEX other_session_interval_idx ON other_session("interval_start");
CREATE INDEX other_session_last_activity_idx ON other_session("last_activity");

-- --------------------
-- aged other session table
-- holds aged non-tcp and non-udp sessions (see age_other_session() to see aging process)
-- --------------------
CREATE TABLE aged_other_session 
(
    id bigserial primary key,  -- unique id for each row
    session_start timestamp with time zone,  -- start time of tcp session
    interval_start timestamp with time zone, -- start time of interval
    hour_start timestamp with time zone, -- start time of hour
    day_start timestamp with time zone, -- start time of day
    week_start timestamp with time zone, -- start time of week
    last_activity timestamp with time zone,  -- last activity seen for this tcp session
    last_interval timestamp with time zone, -- start time of last interval of activity
    last_hour timestamp with time zone, -- start time of last hour of activity
    last_day timestamp with time zone, -- start time of last day of activity
    last_week timestamp with time zone, -- start time of last week of activity
    router character varying(16),  -- IP address of router/exporter
    src_addr inet not null,  -- source IP address
    dst_addr inet not null,  -- destination IP address
    src_port integer default 0,  -- source port number
    dst_port integer default 0,  -- destination port number
    prot integer not null,  -- protocol number
    in_pkts bigint default 0,  -- number of packets entering the src_addr/src_port
    out_pkts bigint default 0,  -- number of packets leaving the src_addr/src_port
    in_bytes bigint default 0,  -- number of bytes entering the src_addr/src_port 
    out_bytes bigint default 0,  -- number of bytes leaving the src_addr/src_port
    last_update timestamp with time zone default now() -- last time this row was updated
) with (fillfactor = 90);
CREATE INDEX aged_other_session_addr_port_idx ON aged_other_session("router", "src_addr", "dst_addr", "src_port", "dst_port", "prot");
CREATE INDEX aged_other_session_interval_idx ON aged_other_session("interval_start");
CREATE INDEX aged_other_session_last_activity_idx ON aged_other_session("last_activity");

--  --------------------
-- internal network table
-- holds internal networks or IP addresses
-- ---------------------- 
CREATE TABLE internal_net
(
   id bigserial primary key, -- unique id for each row
   network cidr not null unique -- network or IP in CIDR notation
);
INSERT INTO internal_net(network) VALUES('192.168.0.0/16');  -- initialize default internal network

-- -----------------------------------------------------------------------------------------
-- TABLE PRIVILEGES
-- -----------------------------------------------------------------------------------------

GRANT insert, select ON flow TO psyche_col;
GRANT usage, select ON flow_id_seq TO psyche_col;
GRANT select ON flow TO psyche_ui;
GRANT usage, select ON flow_id_seq TO psyche_ui;

GRANT select, update ON col_state TO psyche_col;
GRANT select, update ON src_traffic_state TO psyche_col;
GRANT select, update ON dst_traffic_state TO psyche_col;
GRANT select, update ON tcp_session_state TO psyche_col;
GRANT select, update ON udp_session_state TO psyche_col;
GRANT select, update ON other_session_state TO psyche_col;
GRANT select, update ON aged_tcp_session_state TO psyche_col;
GRANT select, update ON aged_udp_session_state TO psyche_col;
GRANT select, update ON aged_other_session_state TO psyche_col;

GRANT select, insert, update ON src_traffic TO psyche_col;
GRANT usage, select ON src_traffic_id_seq TO psyche_col;
GRANT select ON src_traffic TO psyche_ui;
GRANT usage, select ON src_traffic_id_seq TO psyche_ui;

GRANT select, insert, update ON dst_traffic TO psyche_col;
GRANT usage, select ON dst_traffic_id_seq TO psyche_col;
GRANT select ON dst_traffic TO psyche_ui;
GRANT usage, select ON dst_traffic_id_seq TO psyche_ui;

GRANT select, insert, update, delete ON tcp_session TO psyche_col;
GRANT usage, select ON tcp_session_id_seq TO psyche_col;
GRANT select ON tcp_session TO psyche_ui;
GRANT usage, select ON tcp_session_id_seq TO psyche_ui;

GRANT select, insert, update ON aged_tcp_session TO psyche_col;
GRANT usage, select ON aged_tcp_session_id_seq TO psyche_col;
GRANT select ON aged_tcp_session TO psyche_ui;
GRANT usage, select ON aged_tcp_session_id_seq TO psyche_ui;

GRANT select, insert, update, delete ON udp_session TO psyche_col;
GRANT usage, select ON udp_session_id_seq TO psyche_col;
GRANT select ON udp_session TO psyche_ui;
GRANT usage, select ON udp_session_id_seq TO psyche_ui;

GRANT select, insert, update ON aged_udp_session TO psyche_col;
GRANT usage, select ON aged_udp_session_id_seq TO psyche_col;
GRANT select ON aged_udp_session TO psyche_ui;
GRANT usage, select ON aged_udp_session_id_seq TO psyche_ui;

GRANT select, insert, update, delete ON other_session TO psyche_col;
GRANT usage, select ON other_session_id_seq TO psyche_col;
GRANT select ON other_session TO psyche_ui;
GRANT usage, select ON other_session_id_seq TO psyche_ui;

GRANT select, insert, update ON aged_other_session TO psyche_col;
GRANT usage, select ON aged_other_session_id_seq TO psyche_col;
GRANT select ON aged_other_session TO psyche_ui;
GRANT usage, select ON aged_other_session_id_seq TO psyche_ui;

GRANT select, insert, update, delete ON internal_net TO psyche_ui;
GRANT usage, select ON internal_net_id_seq TO psyche_ui;

-- -----------------------------------------------------------------------------------------
-- STORED PROCEDURES
-- -----------------------------------------------------------------------------------------

-- initialize PL/pgSQL
CREATE PROCEDURAL LANGUAGE plpgsql;

-- --------------------
-- this_interval function
--
-- calculate an interval boundary based on input timestamp (with time zone)
-- --------------------
CREATE OR REPLACE FUNCTION this_interval(ts timestamp with time zone, interval_length interval) RETURNS timestamp with time zone AS 
$$
DECLARE
        e integer := FLOOR(EXTRACT(epoch FROM ts::timestamp))::integer;
        l integer := EXTRACT(epoch FROM interval_length)::integer;
BEGIN
        RETURN ('epoch'::timestamp with time zone + (e - e%l) * '1 second'::interval)::timestamp with time zone;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- aggregate_src_traffic function
--
-- retrieves all non-aggregated flows from the flow table and
-- aggregates them into the src_traffic table
-- --------------------
CREATE OR REPLACE FUNCTION aggregate_src_traffic() RETURNS bigint AS 
$$
DECLARE
        _interval_start timestamp with time zone;
        _half_hour_start timestamp with time zone;
        _router varchar(16);
        _addr inet;
        _port integer;
        _prot integer;
        _bytes bigint;
        _pkts bigint;
        _start_fid bigint;
        _last_fid bigint;
BEGIN
        LOCK TABLE src_traffic_state IN ACCESS EXCLUSIVE MODE;
        SELECT last_fid INTO _start_fid FROM src_traffic_state;
	SELECT max(id) INTO _last_fid FROM flow;
        
	_last_fid := GREATEST(0, _last_fid);
        IF _last_fid = 0 THEN
          RETURN 1;
        END IF;

        IF _start_fid = _last_fid THEN
          RETURN _last_fid;
        END IF;

        CREATE TEMPORARY TABLE temp_src_traffic ON COMMIT DROP AS
          SELECT this_interval(start_time, '300') AS interval_start,
		 this_interval(start_time, '1800') AS half_hour_start,
		 router,
                 src_addr AS addr,
                 src_port AS port,
                 prot,
                 sum(bytes) AS bytes,
		 sum(pkts) AS pkts
          FROM flow WHERE id > _start_fid AND id <= _last_fid
          GROUP BY interval_start, half_hour_start, router, addr, port, prot;

        FOR _interval_start, _half_hour_start, _router, _addr, _port, _prot, _bytes, _pkts IN 
          SELECT interval_start, half_hour_start, router, addr, port, prot, bytes, pkts FROM temp_src_traffic
        LOOP
          UPDATE src_traffic
            SET bytes = bytes + _bytes, pkts = pkts + _pkts, last_update = now()
            WHERE interval_start = _interval_start AND
		  router = _router AND
                  addr = _addr AND
                  port = _port AND
                  prot = _prot;
          IF NOT FOUND THEN
            INSERT INTO 
              src_traffic (interval_start, half_hour_start, hour_start, day_start, week_start, 
                           router, addr, port, prot, bytes, pkts, last_update)
            VALUES
              (_interval_start, _half_hour_start, date_trunc('hour', _interval_start), 
               date_trunc('day', _interval_start), date_trunc('week', _interval_start),
               _router, _addr, _port, _prot, _bytes, _pkts, now());
          END IF;
        END LOOP;

        UPDATE src_traffic_state 
          SET 
            last_fid = _last_fid 
          WHERE 
            id = 1;

        RETURN _last_fid;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- aggregate_dst_traffic function
--
-- retrieves all non-aggregated flows from the flow table and
-- aggregates them into the dst_traffic table
-- --------------------
CREATE OR REPLACE FUNCTION aggregate_dst_traffic() RETURNS bigint AS 
$$
DECLARE
        _interval_start timestamp with time zone;
        _half_hour_start timestamp with time zone;
	_router varchar(16);
        _addr inet;
        _port integer;
        _prot integer;
        _bytes bigint;
        _pkts bigint;
        _start_fid bigint;
        _last_fid bigint;
BEGIN
        LOCK TABLE dst_traffic_state IN ACCESS EXCLUSIVE MODE;
        SELECT last_fid INTO _start_fid FROM dst_traffic_state;
	SELECT max(id) INTO _last_fid FROM flow;

        _last_fid := GREATEST(0, _last_fid);
        IF _last_fid = 0 THEN
          RETURN 1;
        END IF;

        IF _start_fid = _last_fid THEN
          RETURN _last_fid;
        END IF;

        CREATE TEMPORARY TABLE temp_dst_traffic ON COMMIT DROP AS
          SELECT this_interval(start_time, '300') AS interval_start,
		 this_interval(start_time, '1800') AS half_hour_start,
		 router,
                 dst_addr AS addr,
                 dst_port AS port,
                 prot,
                 sum(bytes) AS bytes,
                 sum(pkts) AS pkts
          FROM flow WHERE id > _start_fid AND id <= _last_fid
          GROUP BY interval_start, half_hour_start, router, addr, port, prot;

        FOR _interval_start, _half_hour_start, _router, _addr, _port, _prot, _bytes, _pkts IN 
          SELECT interval_start, half_hour_start, router, addr, port, prot, bytes, pkts FROM temp_dst_traffic
        LOOP
          UPDATE dst_traffic
            SET bytes = bytes + _bytes, pkts = pkts + _pkts, last_update = now()
            WHERE interval_start = _interval_start AND
		  router = _router AND
                  addr = _addr AND
                  port = _port AND
                  prot = _prot;
          IF NOT FOUND THEN
            INSERT INTO 
              dst_traffic (interval_start, half_hour_start, hour_start, day_start, week_start, router, addr, port, prot, bytes, pkts, last_update)
            VALUES
              (_interval_start, _half_hour_start, date_trunc('hour', _interval_start), 
               date_trunc('day', _interval_start), date_trunc('week', _interval_start),
               _router, _addr, _port, _prot, _bytes, _pkts, now());
          END IF;
        END LOOP;

        UPDATE dst_traffic_state 
          SET 
            last_fid = _last_fid 
          WHERE 
            id = 1;

        RETURN _last_fid;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- aggregate_tcp_session function
--
-- retrieves all non-aggregated tcp session flows from the flow table and
-- aggregates them into the tcp_session table
-- --------------------
CREATE OR REPLACE FUNCTION aggregate_tcp_session() RETURNS bigint AS 
$$
DECLARE
        _start_time timestamp with time zone;
        _end_time timestamp with time zone;
	_router varchar(16);
        _src_addr inet;
        _dst_addr inet;
        _src_port integer;
        _dst_port integer;
        _pkts bigint;
        _bytes bigint;
        _start_fid bigint;
        _last_fid bigint;
BEGIN
        LOCK TABLE tcp_session_state IN ACCESS EXCLUSIVE MODE;
        SELECT last_fid INTO _start_fid FROM tcp_session_state;
	SELECT max(id) INTO _last_fid FROM flow;

        _last_fid := GREATEST(0, _last_fid);
        IF _last_fid = 0 THEN
          RETURN 1;
        END IF;

        _last_fid := LEAST(_start_fid + 1250, _last_fid);
        IF _start_fid = _last_fid THEN
          RETURN _last_fid;
        END IF;

        FOR _start_time, _end_time, _router, _src_addr, _dst_addr, _src_port, _dst_port, _pkts, _bytes IN 
          SELECT 
            min(start_time), max(end_time), router, src_addr, dst_addr, src_port, dst_port, sum(pkts), sum(bytes)
          FROM
            flow
          WHERE
            prot = 6 AND
            id > _start_fid AND id <= _last_fid
	  GROUP BY
	    router, src_addr, dst_addr, src_port, dst_port
        LOOP
          IF _src_addr < _dst_addr THEN
            UPDATE tcp_session
              SET
                out_pkts = out_pkts + _pkts, 
                out_bytes = out_bytes + _bytes,
                session_start = LEAST( _start_time, session_start ),
	 	interval_start = LEAST( this_interval(_start_time, '300'), interval_start ),
		hour_start = LEAST( date_trunc('hour', _start_time), hour_start ),
		day_start = LEAST( date_trunc('day', _start_time), day_start ),
		week_start = LEAST( date_trunc('week', _start_time), week_start),
                last_activity = GREATEST( _end_time, last_activity ),
	 	last_interval = GREATEST( this_interval(_end_time, '300'), last_interval ),
		last_hour = GREATEST( date_trunc('hour', _end_time), last_hour ),
		last_day = GREATEST( date_trunc('day', _end_time), last_day ),
		last_week = GREATEST( date_trunc('week', _end_time), last_week),
		last_update = now()
              WHERE
		router   = _router   AND
                src_addr = _src_addr AND
                dst_addr = _dst_addr AND
                src_port = _src_port AND
                dst_port = _dst_port;
            IF NOT FOUND THEN
              INSERT INTO 
                tcp_session(interval_start, hour_start, day_start, week_start, 
			    last_interval, last_hour, last_day, last_week,
			    session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, out_pkts, out_bytes, last_update)
              VALUES
                (this_interval(_start_time, '300'), date_trunc('hour', _start_time), date_trunc('day', _start_time), date_trunc('week', _start_time),
                 this_interval(_end_time, '300'), date_trunc('hour', _end_time), date_trunc('day', _end_time), date_trunc('week', _end_time),
                 _start_time, _end_time, _router, _src_addr, _dst_addr, _src_port, _dst_port, _pkts, _bytes, now());
            END IF;
          ELSE
            UPDATE tcp_session
              SET
                in_pkts = in_pkts + _pkts, 
                in_bytes = in_bytes + _bytes,
                session_start = LEAST( _start_time, session_start ),
	 	interval_start = LEAST( this_interval(_start_time, '300'), interval_start ),
		hour_start = LEAST( date_trunc('hour', _start_time), hour_start ),
		day_start = LEAST( date_trunc('day', _start_time), day_start ),
		week_start = LEAST( date_trunc('week', _start_time), week_start),
                last_activity = GREATEST( _end_time, last_activity ),
	 	last_interval = GREATEST( this_interval(_end_time, '300'), last_interval ),
		last_hour = GREATEST( date_trunc('hour', _end_time), last_hour ),
		last_day = GREATEST( date_trunc('day', _end_time), last_day ),
		last_week = GREATEST( date_trunc('week', _end_time), last_week),
		last_update = now()
              WHERE
		router   = _router   AND
                src_addr = _dst_addr AND
                dst_addr = _src_addr AND
                src_port = _dst_port AND
                dst_port = _src_port;
            IF NOT FOUND THEN
              INSERT INTO 
                tcp_session(interval_start, hour_start, day_start, week_start, 
			    last_interval, last_hour, last_day, last_week,
                	    session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, in_pkts, in_bytes, last_update)
              VALUES
                (this_interval(_start_time, '300'), date_trunc('hour', _start_time), date_trunc('day', _start_time), date_trunc('week', _start_time),
                 this_interval(_end_time, '300'), date_trunc('hour', _end_time), date_trunc('day', _end_time), date_trunc('week', _end_time),
                 _start_time, _end_time, _router, _dst_addr, _src_addr, _dst_port, _src_port, _pkts, _bytes, now());
            END IF;
          END IF;
        END LOOP;

        UPDATE tcp_session_state 
          SET
            last_fid = _last_fid
          WHERE 
            id = 1;

        RETURN _last_fid;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- age_tcp_session function
--
-- moves old sessions from tcp_session to aged_tcp_session
-- --------------------
CREATE OR REPLACE FUNCTION age_tcp_session() RETURNS void AS
$$
DECLARE 
	_time timestamp with time zone;
	_interval interval;
BEGIN
        LOCK TABLE aged_tcp_session_state IN ACCESS EXCLUSIVE MODE;
	SELECT tcp_age_interval INTO _interval FROM col_state;
 
	_time := now() - _interval;	

        CREATE TEMPORARY TABLE temp_tcp_session ON COMMIT DROP AS
          SELECT 
	    *
          FROM 
	    tcp_session
	  WHERE
            last_activity < _time
	  FOR UPDATE;

        DELETE FROM
          tcp_session
        WHERE
          last_activity < _time;

	INSERT INTO
	  aged_tcp_session
	SELECT
	  *
	FROM
	  temp_tcp_session;

	RETURN;
END;
$$
LANGUAGE plpgsql;
	          
-- --------------------
-- aggregate_udp_session function
--
-- retrieves all non-aggregated udp session flows from the flow table and
-- aggregates them into the udp_session table
-- --------------------
CREATE OR REPLACE FUNCTION aggregate_udp_session() RETURNS bigint AS 
$$
DECLARE
        _start_time timestamp with time zone;
        _end_time timestamp with time zone;
	_router varchar(16);
        _src_addr inet;
        _dst_addr inet;
        _src_port integer;
        _dst_port integer;
        _pkts bigint;
        _bytes bigint;
        _start_fid bigint;
        _last_fid bigint;
BEGIN
        LOCK TABLE udp_session_state IN ACCESS EXCLUSIVE MODE;
        SELECT last_fid INTO _start_fid FROM udp_session_state;
	SELECT max(id) INTO _last_fid FROM flow;

        _last_fid := GREATEST(0, _last_fid);
        IF _last_fid = 0 THEN
          RETURN 1;
        END IF;

        _last_fid := LEAST(_start_fid + 1250, _last_fid);
        IF _start_fid = _last_fid THEN
          RETURN _last_fid;
        END IF;

        FOR _start_time, _end_time, _router, _src_addr, _dst_addr, _src_port, _dst_port, _pkts, _bytes IN 
          SELECT 
            min(start_time), max(end_time), router, src_addr, dst_addr, src_port, dst_port, sum(pkts), sum(bytes)
          FROM
            flow
          WHERE
            prot = 17 AND
            id > _start_fid AND id <= _last_fid
	  GROUP BY
	    router, src_addr, dst_addr, src_port, dst_port
        LOOP
          IF _src_addr < _dst_addr THEN
            UPDATE udp_session
              SET
                out_pkts = out_pkts + _pkts, 
                out_bytes = out_bytes + _bytes,
                session_start = LEAST( _start_time, session_start ),
	 	interval_start = LEAST( this_interval(_start_time, '300'), interval_start ),
		hour_start = LEAST( date_trunc('hour', _start_time), hour_start ),
		day_start = LEAST( date_trunc('day', _start_time), day_start ),
		week_start = LEAST( date_trunc('week', _start_time), week_start),
                last_activity = GREATEST( _end_time, last_activity ),
	 	last_interval = GREATEST( this_interval(_end_time, '300'), last_interval ),
		last_hour = GREATEST( date_trunc('hour', _end_time), last_hour ),
		last_day = GREATEST( date_trunc('day', _end_time), last_day ),
		last_week = GREATEST( date_trunc('week', _end_time), last_week),
		last_update = now()
              WHERE
		router   = _router   AND
                src_addr = _src_addr AND
                dst_addr = _dst_addr AND
                src_port = _src_port AND
                dst_port = _dst_port;
            IF NOT FOUND THEN
              INSERT INTO 
                udp_session(interval_start, hour_start, day_start, week_start, 
			    last_interval, last_hour, last_day, last_week,
			    session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, out_pkts, out_bytes, last_update)
              VALUES
                (this_interval(_start_time, '300'), date_trunc('hour', _start_time), date_trunc('day', _start_time), date_trunc('week', _start_time),
                 this_interval(_end_time, '300'), date_trunc('hour', _end_time), date_trunc('day', _end_time), date_trunc('week', _end_time),
                 _start_time, _end_time, _router, _src_addr, _dst_addr, _src_port, _dst_port, _pkts, _bytes, now());
            END IF;
          ELSE
            UPDATE udp_session
              SET
                in_pkts = in_pkts + _pkts, 
                in_bytes = in_bytes + _bytes,
                session_start = LEAST( _start_time, session_start ),
	 	interval_start = LEAST( this_interval(_start_time, '300'), interval_start ),
		hour_start = LEAST( date_trunc('hour', _start_time), hour_start ),
		day_start = LEAST( date_trunc('day', _start_time), day_start ),
		week_start = LEAST( date_trunc('week', _start_time), week_start),
                last_activity = GREATEST( _end_time, last_activity ),
	 	last_interval = GREATEST( this_interval(_end_time, '300'), last_interval ),
		last_hour = GREATEST( date_trunc('hour', _end_time), last_hour ),
		last_day = GREATEST( date_trunc('day', _end_time), last_day ),
		last_week = GREATEST( date_trunc('week', _end_time), last_week),
		last_update = now()
              WHERE
		router   = _router   AND
                src_addr = _dst_addr AND
                dst_addr = _src_addr AND
                src_port = _dst_port AND
                dst_port = _src_port;
            IF NOT FOUND THEN
              INSERT INTO 
                udp_session(interval_start, hour_start, day_start, week_start, 
			    last_interval, last_hour, last_day, last_week,
                	    session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, in_pkts, in_bytes, last_update)
              VALUES
                (this_interval(_start_time, '300'), date_trunc('hour', _start_time), date_trunc('day', _start_time), date_trunc('week', _start_time),
                 this_interval(_end_time, '300'), date_trunc('hour', _end_time), date_trunc('day', _end_time), date_trunc('week', _end_time),
                 _start_time, _end_time, _router, _dst_addr, _src_addr, _dst_port, _src_port, _pkts, _bytes, now());
            END IF;
          END IF;
        END LOOP;

        UPDATE udp_session_state 
          SET
            last_fid = _last_fid
          WHERE 
            id = 1;

        RETURN _last_fid;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- age_udp_session function
--
-- moves old sessions from udp_session to aged_udp_session
-- --------------------
CREATE OR REPLACE FUNCTION age_udp_session() RETURNS void AS
$$
DECLARE 
	_time timestamp with time zone;
	_interval interval;
BEGIN
        LOCK TABLE aged_udp_session_state IN ACCESS EXCLUSIVE MODE;
	SELECT udp_age_interval INTO _interval FROM col_state;
 
	_time := now() - _interval;	

        CREATE TEMPORARY TABLE temp_udp_session ON COMMIT DROP AS
          SELECT 
	    *
          FROM 
	    udp_session
	  WHERE
            last_activity < _time
	  FOR UPDATE;

        DELETE FROM
          udp_session
        WHERE
          last_activity < _time;

	INSERT INTO
	  aged_udp_session
	SELECT
	  *
	FROM
	  temp_udp_session;

	RETURN;
END;
$$
LANGUAGE plpgsql;
	          
-- --------------------
-- aggregate_other_session function
--
-- retrieves all non-aggregated non-tcp and non-udp session flows from the flow table and
-- aggregates them into the other_session table
-- --------------------
CREATE OR REPLACE FUNCTION aggregate_other_session() RETURNS bigint AS 
$$
DECLARE
        _start_time timestamp with time zone;
        _end_time timestamp with time zone;
	_router varchar(16);
        _prot integer;
        _src_addr inet;
        _dst_addr inet;
        _src_port integer;
        _dst_port integer;
        _pkts bigint;
        _bytes bigint;
        _start_fid bigint;
        _last_fid bigint;
BEGIN
        LOCK TABLE other_session_state IN ACCESS EXCLUSIVE MODE;
        SELECT last_fid INTO _start_fid FROM other_session_state;
	SELECT max(id) INTO _last_fid FROM flow;

        _last_fid := GREATEST(0, _last_fid);
        IF _last_fid = 0 THEN
          RETURN 1;
        END IF;

        _last_fid := LEAST(_start_fid + 1250, _last_fid);
        IF _start_fid = _last_fid THEN
          RETURN _last_fid;
        END IF;

        FOR _start_time, _end_time, _router, _prot, _src_addr, _dst_addr, _src_port, _dst_port, _pkts, _bytes IN 
          SELECT 
            min(start_time), max(end_time), router, prot, src_addr, dst_addr, src_port, dst_port, sum(pkts), sum(bytes)
          FROM
            flow
          WHERE
            prot != 6 AND prot != 17 AND
            id > _start_fid AND id <= _last_fid
	  GROUP BY
	    router, prot, src_addr, dst_addr, src_port, dst_port
        LOOP
          IF _src_addr < _dst_addr THEN
            UPDATE other_session
              SET
                out_pkts = out_pkts + _pkts, 
                out_bytes = out_bytes + _bytes,
                session_start = LEAST( _start_time, session_start ),
	 	interval_start = LEAST( this_interval(_start_time, '300'), interval_start ),
		hour_start = LEAST( date_trunc('hour', _start_time), hour_start ),
		day_start = LEAST( date_trunc('day', _start_time), day_start ),
		week_start = LEAST( date_trunc('week', _start_time), week_start),
                last_activity = GREATEST( _end_time, last_activity ),
	 	last_interval = GREATEST( this_interval(_end_time, '300'), last_interval ),
		last_hour = GREATEST( date_trunc('hour', _end_time), last_hour ),
		last_day = GREATEST( date_trunc('day', _end_time), last_day ),
		last_week = GREATEST( date_trunc('week', _end_time), last_week),
		last_update = now()
              WHERE
		router   = _router   AND
		prot     = _prot     AND
                src_addr = _src_addr AND
                dst_addr = _dst_addr AND
                src_port = _src_port AND
                dst_port = _dst_port;
            IF NOT FOUND THEN
              INSERT INTO 
                other_session(interval_start, hour_start, day_start, week_start, 
		  	      last_interval, last_hour, last_day, last_week,
			      session_start, last_activity, router, prot, src_addr, dst_addr, src_port, dst_port, out_pkts, out_bytes, last_update)
              VALUES
                (this_interval(_start_time, '300'), date_trunc('hour', _start_time), date_trunc('day', _start_time), date_trunc('week', _start_time),
                 this_interval(_end_time, '300'), date_trunc('hour', _end_time), date_trunc('day', _end_time), date_trunc('week', _end_time),
                 _start_time, _end_time, _router, _prot, _src_addr, _dst_addr, _src_port, _dst_port, _pkts, _bytes, now());
            END IF;
          ELSE
            UPDATE other_session
              SET
                in_pkts = in_pkts + _pkts, 
                in_bytes = in_bytes + _bytes,
                session_start = LEAST( _start_time, session_start ),
	 	interval_start = LEAST( this_interval(_start_time, '300'), interval_start ),
		hour_start = LEAST( date_trunc('hour', _start_time), hour_start ),
		day_start = LEAST( date_trunc('day', _start_time), day_start ),
		week_start = LEAST( date_trunc('week', _start_time), week_start),
                last_activity = GREATEST( _end_time, last_activity ),
	 	last_interval = GREATEST( this_interval(_end_time, '300'), last_interval ),
		last_hour = GREATEST( date_trunc('hour', _end_time), last_hour ),
		last_day = GREATEST( date_trunc('day', _end_time), last_day ),
		last_week = GREATEST( date_trunc('week', _end_time), last_week),
		last_update = now()
              WHERE
		router   = _router   AND
		prot     = _prot     AND
                src_addr = _dst_addr AND
                dst_addr = _src_addr AND
                src_port = _dst_port AND
                dst_port = _src_port;
            IF NOT FOUND THEN
              INSERT INTO 
                other_session(interval_start, hour_start, day_start, week_start, 
			      last_interval, last_hour, last_day, last_week,
                	      session_start, last_activity, router, prot, src_addr, dst_addr, src_port, dst_port, in_pkts, in_bytes, last_update)
              VALUES
                (this_interval(_start_time, '300'), date_trunc('hour', _start_time), date_trunc('day', _start_time), date_trunc('week', _start_time),
                 this_interval(_end_time, '300'), date_trunc('hour', _end_time), date_trunc('day', _end_time), date_trunc('week', _end_time),
                 _start_time, _end_time, _router, _prot, _dst_addr, _src_addr, _dst_port, _src_port, _pkts, _bytes, now());
            END IF;
          END IF;
        END LOOP;

        UPDATE other_session_state 
          SET
            last_fid = _last_fid
          WHERE 
            id = 1;

        RETURN _last_fid;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- age_other_session function
--
-- moves old sessions from other_session to aged_other_session
-- --------------------
CREATE OR REPLACE FUNCTION age_other_session() RETURNS void AS
$$
DECLARE 
	_time timestamp with time zone;
	_interval interval;
BEGIN
        LOCK TABLE aged_other_session_state IN ACCESS EXCLUSIVE MODE;
	SELECT other_age_interval INTO _interval FROM col_state;
 
	_time := now() - _interval;	

        CREATE TEMPORARY TABLE temp_other_session ON COMMIT DROP AS
          SELECT 
	    *
          FROM 
	    other_session
	  WHERE
            last_activity < _time
	  FOR UPDATE;

        DELETE FROM
          other_session
        WHERE
          last_activity < _time;

	INSERT INTO
	  aged_other_session
	SELECT
	  *
	FROM
	  temp_other_session;

	RETURN;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_time_clause function
--
-- creates time where clause 
-- --------------------
CREATE OR REPLACE FUNCTION get_time_clause(varchar(16), timestamp with time zone, timestamp with time zone) RETURNS text AS 
$$
DECLARE
	_column	     ALIAS FOR $1;
        _start_time  ALIAS FOR $2;
        _end_time    ALIAS FOR $3;
	clause text;
BEGIN
  clause := ' ' || quote_ident(_column) || ' BETWEEN ' || quote_literal(_start_time) || '::timestamp with time zone AND '
                                        || quote_literal(_end_time) || '::timestamp with time zone ';
  RETURN clause;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_net_clause function
--
-- creates network where clause 
-- --------------------
CREATE OR REPLACE FUNCTION get_net_clause(int, varchar(18)) RETURNS text AS 
$$
DECLARE
	_addr_flag   ALIAS FOR $1; -- 0 = column is addr; 1 = columns are src_addr and dst_addr
        _network     ALIAS FOR $2;
	clause text := ' ';
BEGIN
  IF _network != 'none' THEN
    IF _addr_flag = 0 THEN
      clause := ' AND addr <<= ' || quote_literal(_network) || '::cidr ';
    ELSIF _addr_flag = 1 THEN
      clause := ' AND (src_addr <<= ' || quote_literal(_network) || '::cidr OR dst_addr <<= ' || quote_literal(_network) || '::cidr) ';
    END IF;
  END IF;

  RETURN clause;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_net_clause function
--
-- creates network where clause 
-- --------------------
CREATE OR REPLACE FUNCTION get_net_clause(varchar(18), varchar(18)) RETURNS text AS 
$$
DECLARE
        _src_network     ALIAS FOR $1;
        _dst_network     ALIAS FOR $2;
	clause text := ' ';
BEGIN
  IF _src_network != 'none' THEN
      IF _dst_network != 'none' THEN
          clause := ' AND (src_addr <<= ' || quote_literal(_src_network) || '::cidr AND dst_addr <<= ' || quote_literal(_dst_network) || '::cidr) ';
      ELSE
          clause := ' AND (src_addr <<= ' || quote_literal(_src_network) || '::cidr) ';
      END IF;
  ELSIF _dst_network != 'none' THEN
      clause := ' AND (dst_addr <<= ' || quote_literal(_dst_network) || '::cidr) ';
  END IF;

  RETURN clause;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_router_clause function
--
-- creates router where clause 
-- --------------------
CREATE OR REPLACE FUNCTION get_router_clause(varchar(16)) RETURNS text AS 
$$
DECLARE
        _router     ALIAS FOR $1;
	clause text := ' ';
BEGIN
  IF _router != 'none' THEN
    clause := ' AND router = ' || quote_literal(_router) || '::varchar ';
  END IF;

  RETURN clause;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_prot_clause function
--
-- creates protocol where clause 
-- --------------------
CREATE OR REPLACE FUNCTION get_prot_clause(integer[][], integer[], integer[][], integer[]) RETURNS text AS 
$$
DECLARE
	_iprot_range ALIAS FOR $1;
	_iprot       ALIAS FOR $2;
	_eprot_range ALIAS FOR $3;
	_eprot       ALIAS FOR $4;
	clause text := ' ';
	i integer := 1;
BEGIN
  WHILE _iprot_range[i:i] != '{}'::integer[] LOOP
    IF i = 1 THEN
      clause := clause || ' AND (';
    ELSE
      clause := clause || ' OR ';
    END IF;
    clause := clause || 'prot BETWEEN ' || quote_literal(_iprot_range[i][1]) || ' AND ' || quote_literal(_iprot_range[i][2]) || ' ';
    i := i + 1;
  END LOOP;

  IF _iprot[1] IS NOT NULL THEN
    IF i = 1 THEN
      clause := clause || ' AND ';
    ELSE
      clause := clause || ' OR ';
    END IF;
    clause := clause || 'prot = ANY (' || quote_literal(_iprot) || '::integer[]) ';
  END IF;
  IF i > 1 THEN
    clause := clause || ')';
  END IF;

  i := 1;
  WHILE _eprot_range[i:i] != '{}'::integer[] LOOP
    clause := clause || ' AND NOT prot BETWEEN ' || quote_literal(_eprot_range[i][1]) || ' AND ' || quote_literal(_eprot_range[i][2]) || ' ';
    i := i + 1;
  END LOOP;

  IF _eprot[1] IS NOT NULL THEN
    clause := clause || ' AND NOT prot = ANY (' || quote_literal(_eprot) || '::integer[]) ';
  END IF;

  RETURN clause;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_port_clause function
--
-- creates port where clause 
-- --------------------
CREATE OR REPLACE FUNCTION get_port_clause(int, integer[][], integer[], integer[][], integer[]) RETURNS text AS 
$$
DECLARE
	_port_flag   ALIAS FOR $1; -- 0 = column is port; 1 = columns are src_port and dst_port
	_iport_range ALIAS FOR $2;
	_iport       ALIAS FOR $3;
	_eport_range ALIAS FOR $4;
	_eport       ALIAS FOR $5;
	clause text := ' ';
	i integer := 1;
BEGIN
  WHILE _iport_range[i:i] != '{}'::integer[] LOOP
    IF i = 1 THEN
      clause := clause || ' AND (';
    ELSE
      clause := clause || ' OR ';
    END IF;
    IF _port_flag = 0 THEN 
      clause := clause || 'port BETWEEN ' || quote_literal(_iport_range[i][1]) || ' AND ' || quote_literal(_iport_range[i][2]) || ' ';
    ELSIF _port_flag = 1 THEN
      clause := clause || '(src_port BETWEEN ' || quote_literal(_iport_range[i][1]) || ' AND ' || quote_literal(_iport_range[i][2]) || ' OR
	                    dst_port BETWEEN ' || quote_literal(_iport_range[i][1]) || ' AND ' || quote_literal(_iport_range[i][2]) || ') ';
    END IF;
    i := i + 1;
  END LOOP;

  IF _iport[1] IS NOT NULL THEN
    IF i = 1 THEN
      clause := clause || ' AND ';
    ELSE
      clause := clause || ' OR ';
    END IF;
    IF _port_flag = 0 THEN 
      clause := clause || 'port = ANY (' || quote_literal(_iport) || '::integer[]) ';
    ELSIF _port_flag = 1 THEN
      clause := clause || '(src_port = ANY (' || quote_literal(_iport) || '::integer[]) OR
                            dst_port = ANY (' || quote_literal(_iport) || '::integer[])) ';
    END IF;
  END IF;
  IF i > 1 THEN
    clause := clause || ')';
  END IF;

  i := 1;
  WHILE _eport_range[i:i] != '{}'::integer[] LOOP
    IF _port_flag = 0 THEN 
      clause := clause || ' AND NOT port BETWEEN ' || quote_literal(_eport_range[i][1]) || ' AND ' || quote_literal(_eport_range[i][2]) || ' ';
    ELSIF _port_flag = 1 THEN
      clause := clause || ' AND NOT src_port BETWEEN ' || quote_literal(_eport_range[i][1]) || ' AND ' || quote_literal(_eport_range[i][2]) ||
                          ' AND NOT dst_port BETWEEN ' || quote_literal(_eport_range[i][1]) || ' AND ' || quote_literal(_eport_range[i][2]) || ' ';
    END IF;
    i := i + 1;
  END LOOP;

  IF _eport[1] IS NOT NULL THEN
    IF _port_flag = 0 THEN 
      clause := clause || ' AND NOT port = ANY (' || quote_literal(_eport) || '::integer[]) ';
    ELSIF _port_flag = 1 THEN
      clause := clause || ' AND NOT src_port = ANY (' || quote_literal(_eport) || '::integer[]) 
	                    AND NOT dst_port = ANY (' || quote_literal(_eport) || '::integer[]) ';
    END IF;
  END IF;

  RETURN clause;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_div_clause function
--
-- creates division by zero where clause 
-- --------------------
CREATE OR REPLACE FUNCTION get_div_clause() RETURNS text AS 
$$
DECLARE
	clause text := ' ';
BEGIN
  clause := ' AND in_bytes > 0 AND out_bytes > 0 ';

  RETURN clause;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_top_talkers function
--
-- retrieves the bandwidth used by the top X ports or hosts per interval
-- --------------------
CREATE OR REPLACE FUNCTION get_top_talkers(varchar(18), varchar(16), varchar(3), 
                                           integer[][], integer[], integer[][], integer[], 
                                           integer[][], integer[], integer[][], integer[], 
                                           integer, varchar(4), timestamp with time zone, timestamp with time zone, 
				           varchar(12), OUT bigint, OUT refcursor) AS
$$
DECLARE
	_network     ALIAS FOR $1;
	_router      ALIAS FOR $2;
	_dir         ALIAS FOR $3;
	_iprot_range ALIAS FOR $4;
	_iprot       ALIAS FOR $5;
	_eprot_range ALIAS FOR $6;
	_eprot       ALIAS FOR $7;
	_iport_range ALIAS FOR $8;
	_iport       ALIAS FOR $9;
	_eport_range ALIAS FOR $10;
	_eport       ALIAS FOR $11;
        _num_talkers ALIAS FOR $12;
        _talker_type ALIAS FOR $13;
        _start_time  ALIAS FOR $14;
        _end_time    ALIAS FOR $15;
        _interval    ALIAS FOR $16;
        _count       ALIAS FOR $17;
        _results     ALIAS FOR $18;
	table_name   text;
	column_name  text;
	interval_name text;
        time_clause  text;
        net_clause   text := ' ';
        router_clause text := ' ';
        prot_clause  text := ' ';
        port_clause  text := ' ';
        query text;
	i integer := 1;
        _interval_start timestamp with time zone;
	_rand integer;
BEGIN
	_count := 0; -- set default count in case function returns without a result set
	SELECT (random() * 1000000)::integer INTO _rand; -- to ensure temp tables are uniquely named

	-- check the direction of traffic to determine the table to query
	IF _dir = 'src' THEN
	  table_name = 'src_traffic';
	ELSIF _dir = 'dst' THEN
	  table_name = 'dst_traffic';
	ELSE
	  RETURN;
	END IF;

	-- check the talker type to determine the column to retrieve
	IF _talker_type = 'port' THEN
	  column_name = 'port';
	ELSIF _talker_type = 'host' THEN
	  column_name = 'addr';
	ELSE
	  RETURN;
	END IF;

	-- check the interval to determine the column to retrieve
	IF _interval = '4 hours' THEN
	  interval_name = 'interval_start';
	ELSIF _interval = '1 day' THEN
	  interval_name = 'half_hour_start';
	ELSIF _interval = '2 days' THEN
	  interval_name = 'hour_start';
	ELSE
	  RETURN;
	END IF;

	-- sanity check the desired number of talkers
	IF _num_talkers < 1 THEN
	  RETURN;
	END IF;

	-- based on the input parameters, assemble the where clauses used in the queries below
	SELECT * INTO time_clause   FROM get_time_clause(interval_name, _start_time, _end_time);
	SELECT * INTO net_clause    FROM get_net_clause(0, _network);
	SELECT * INTO router_clause FROM get_router_clause(_router);
	SELECT * INTO prot_clause   FROM get_prot_clause(_iprot_range, _iprot, _eprot_range, _eprot);
	SELECT * INTO port_clause   FROM get_port_clause(0, _iport_range, _iport, _eport_range, _eport);

	-- Get the top n talkers over the entire interval first
        query := 'SELECT ' || quote_ident(column_name) || ', sum(bytes) as bw
	          FROM ' || quote_ident(table_name) || '
	          WHERE ' || time_clause || net_clause || router_clause || prot_clause || port_clause || '
                  GROUP BY ' || quote_ident(column_name) || '
                  ORDER BY bw DESC
		  LIMIT ' || _num_talkers;
        EXECUTE 'CREATE TEMP TABLE temp_talkers' || _rand || ' ON COMMIT DROP AS ' || query;

	-- Get total bandwidth per interval
        query := 'SELECT ' || quote_ident(interval_name) || ', sum(bytes) as bw
	          FROM ' || quote_ident(table_name) || '
	          WHERE ' || time_clause || net_clause || router_clause || '
                  GROUP BY ' || quote_ident(interval_name) || ' 
                  ORDER BY ' || quote_ident(interval_name);
        EXECUTE 'CREATE TEMP TABLE temp_bw' || _rand || ' ON COMMIT DROP AS ' || query;

	-- Now grab the interval values for the top talkers found above
        query := 'SELECT ' || quote_ident(interval_name) || ', ' || quote_ident(column_name) || ', sum(bytes) as bw
	          FROM ' || quote_ident(table_name) || '
	          WHERE ' || quote_ident(interval_name) || ' IN (SELECT ' || quote_ident(interval_name) || ' FROM temp_bw' || _rand || ')' 
		          || net_clause || router_clause || prot_clause || port_clause || ' AND ' ||
			  quote_ident(column_name) || ' IN (SELECT ' || quote_ident(column_name) || ' 
                                                            FROM temp_talkers' || _rand || ') 
                  GROUP BY ' || quote_ident(interval_name) || ', ' || quote_ident(column_name) || '
                  ORDER BY ' || quote_ident(interval_name) || ', bw DESC';
        EXECUTE 'CREATE TEMP TABLE temp_talkers2' || _rand || ' ON COMMIT DROP AS ' || query;

        --  Finally, add in the 'other' talker values
	table_name := 'temp_talkers2' || _rand;
        query := 'SELECT * FROM 
	         ((SELECT a.' || quote_ident(interval_name) || ' as interval_start, ';
			IF column_name = 'port' THEN
			  query := query || '-1 as ' || quote_ident(column_name) || ', '; 
			ELSE 
			  query := query || quote_literal('0.0.0.0') || '::inet as ' || quote_ident(column_name) || ', '; 
			END IF;
	query := query || ' sum(a.bw - GREATEST(0,b.bw)) as bw
	           FROM temp_bw' || _rand || ' a LEFT OUTER JOIN (SELECT ' || quote_ident(interval_name) || ', sum(bw) as bw
		 			                          FROM ' || quote_ident(table_name) || ' 
					                          GROUP BY ' || quote_ident(interval_name) || ') b 
		   ON a.' || quote_ident(interval_name) || ' = b.' || quote_ident(interval_name) || '
                   GROUP BY a.' || quote_ident(interval_name) || ', ' || quote_ident(column_name) || '
                   ORDER BY a.' || quote_ident(interval_name) || ')
		   UNION
		   (SELECT * FROM ' || quote_ident(table_name) || ')) c
		   ORDER BY c.interval_start, bw DESC';
        EXECUTE 'CREATE TEMP TABLE top_talkers_results' || _rand || ' ON COMMIT DROP AS ' || query;

	table_name := 'top_talkers_results' || _rand;
        EXECUTE 'SELECT count(*) FROM ' || quote_ident(table_name) INTO _count;
        OPEN _results FOR EXECUTE 'SELECT * FROM ' || quote_ident(table_name) || ' ORDER BY interval_start, bw DESC';
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- bw_ratio function
--
-- retrieves the ratio of inbound vs outbound bytes of aggregated sessions
-- --------------------
CREATE OR REPLACE FUNCTION bw_ratio(varchar(18), varchar(16), 
                                    integer[][], integer[], integer[][], integer[], 
                                    integer[][], integer[], integer[][], integer[], 
                                    integer, real, real, timestamp with time zone, timestamp with time zone, 
		      	            OUT bigint, OUT refcursor) AS
$$
DECLARE
	_network     ALIAS FOR $1;
	_router      ALIAS FOR $2;
	_iprot_range ALIAS FOR $3;
	_iprot       ALIAS FOR $4;
	_eprot_range ALIAS FOR $5;
	_eprot       ALIAS FOR $6;
	_iport_range ALIAS FOR $7;
	_iport       ALIAS FOR $8;
	_eport_range ALIAS FOR $9;
	_eport       ALIAS FOR $10;
	_num_buckets ALIAS FOR $11;
	_min_ratio   ALIAS FOR $12;
	_max_ratio   ALIAS FOR $13;
        _start_time  ALIAS FOR $14;
        _end_time    ALIAS FOR $15;
        _count       ALIAS FOR $16;
        _results     ALIAS FOR $17;
        time_clause  text;
        net_clause   text := ' ';
        router_clause text := ' ';
        prot_clause  text := ' ';
        port_clause  text := ' ';
        div_clause   text;
        query text;
	ratio_range real;
	i integer := 1;
        _interval_start timestamp with time zone;
	_ratio real;
	_index integer;
	_min_index integer;
	_max_index integer;
BEGIN
	_count := 0; -- set default count in case function returns without a result set

	-- sanity check the desired number of data buckets
	IF _num_buckets < 1 THEN
	  RETURN;
	END IF;

	-- sanity check the min and max ratios
	IF _min_ratio <= 0 THEN
	  RETURN;
	ELSIF _max_ratio <= _min_ratio THEN
	  RETURN;
	END IF;

	-- based on the input parameters, assemble the where clauses used in the queries below
	SELECT * INTO time_clause   FROM get_time_clause('interval_start', _start_time, _end_time);
	SELECT * INTO net_clause    FROM get_net_clause(1, _network);
	SELECT * INTO router_clause FROM get_router_clause(_router);
	SELECT * INTO prot_clause   FROM get_prot_clause(_iprot_range, _iprot, _eprot_range, _eprot);
	SELECT * INTO port_clause   FROM get_port_clause(1, _iport_range, _iport, _eport_range, _eport);
	SELECT * INTO div_clause    FROM get_div_clause();

	-- Find the bandwidth ratios of the sessions that match the input criteria
        query := 'SELECT ratio FROM 
		  ((SELECT 6 as prot, (in_bytes::real / out_bytes::real) as ratio
	            FROM tcp_session
	            WHERE ' || time_clause || net_clause || router_clause || port_clause || div_clause || ')
		   UNION ALL	
		   (SELECT 6 as prot, (in_bytes::real / out_bytes::real) as ratio
	            FROM aged_tcp_session
	            WHERE ' || time_clause || net_clause || router_clause || port_clause || div_clause || ')) a';
	IF prot_clause != ' ' THEN
	  query := query || ' WHERE 1=1 ' || prot_clause;
	END IF;
        EXECUTE 'CREATE TEMP TABLE temp_tcp ON COMMIT DROP AS ' || query;

        query := 'SELECT ratio FROM 
		  ((SELECT 17 as prot, (in_bytes::real / out_bytes::real) as ratio
	            FROM udp_session
	            WHERE ' || time_clause || net_clause || router_clause || port_clause || div_clause || ')
		   UNION ALL	
		   (SELECT 17 as prot, (in_bytes::real / out_bytes::real) as ratio
	            FROM aged_udp_session
	            WHERE ' || time_clause || net_clause || router_clause || port_clause || div_clause || ')) a';
	IF prot_clause != ' ' THEN
	  query := query || ' WHERE 1=1 ' || prot_clause;
	END IF;
        EXECUTE 'CREATE TEMP TABLE temp_udp ON COMMIT DROP AS ' || query;

        query := 'SELECT ratio FROM 
		  ((SELECT (in_bytes::real / out_bytes::real) as ratio
	            FROM other_session
	            WHERE ' || time_clause || net_clause || router_clause || prot_clause || port_clause || div_clause || ')
		   UNION ALL	
		   (SELECT (in_bytes::real / out_bytes::real) as ratio
	            FROM aged_other_session
	            WHERE ' || time_clause || net_clause || router_clause || prot_clause || port_clause || div_clause || ')) a';
        EXECUTE 'CREATE TEMP TABLE temp_other ON COMMIT DROP AS ' || query;

	query := 'SELECT * FROM (SELECT * FROM temp_tcp UNION ALL SELECT * FROM temp_udp UNION ALL SELECT * FROM temp_other) a 
                  WHERE ratio BETWEEN ' || quote_literal(_min_ratio) || ' AND ' || quote_literal(_max_ratio) || ' ORDER BY ratio';  
        EXECUTE 'CREATE TEMP TABLE bw_ratio_results ON COMMIT DROP AS ' || query;

        SELECT count(*) INTO _count FROM bw_ratio_results;

	-- create the buckets of bandwidth ratios
	IF _count > 0 THEN
	  CREATE TEMP TABLE temp_histo (
	      idx real,
	      ratio_count real
	  ) ON COMMIT DROP;
	  CREATE INDEX temp_histo_idx ON temp_histo("idx");

	  -- initialize the buckets
	  FOR i IN 0.._num_buckets LOOP
	    INSERT INTO temp_histo VALUES(i, 0);
	  END LOOP;

	  -- now fill the buckets  
	  ratio_range := log(_max_ratio) - log(_min_ratio);
	  FOR _ratio IN
	    SELECT ratio FROM bw_ratio_results
	  LOOP
	    -- calculate the bucket index for this ratio
	    _index = floor( (log(_ratio) - log(_min_ratio)) / (ratio_range / _num_buckets) );
	    
	    -- increment the ratio_count of the appropriate bucket
	    UPDATE temp_histo 
	      SET ratio_count = ratio_count + 1 
	      WHERE idx = _index; 
	  END LOOP;	  

	  -- convert the ratio_count values (y-axis) to a log scale
	  FOR _index, _ratio IN
	    SELECT idx, ratio_count FROM temp_histo WHERE ratio_count > 0
	  LOOP
	    UPDATE temp_histo
	      SET ratio_count = log(_ratio)
	      WHERE idx = _index; 
	  END LOOP;

	  -- convert the indices to their ratio values
	  UPDATE temp_histo
	    SET idx = log(_min_ratio) + (idx * ((log(_max_ratio) - log(_min_ratio)) / _num_buckets));

          SELECT count(*) INTO _count FROM temp_histo;
          OPEN _results FOR SELECT * FROM temp_histo ORDER BY idx;
 	END IF;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- bw_by_interval function
--
-- retrieves total bandwidth used (per interval) within the specified timeframe
-- --------------------
CREATE OR REPLACE FUNCTION bw_by_interval(varchar(18), timestamp with time zone, timestamp with time zone, varchar(6), varchar(3), OUT bigint, OUT refcursor) AS
$$
DECLARE
        _network    ALIAS FOR $1;
        _start_time ALIAS FOR $2;
        _end_time   ALIAS FOR $3;
        _interval   ALIAS FOR $4;
        _direction  ALIAS FOR $5;
        _count      ALIAS FOR $6;
        _results    ALIAS FOR $7;
	net         cidr;
	secs integer := 1;
        query text;
        col text;
	i int;
	table_name text;
BEGIN
	IF _interval = 'hour' THEN
	    col := 'interval_start';
	    secs := 300;
	ELSIF _interval = 'day' THEN
	    col := 'interval_start';	
	    secs := 300;
	ELSE -- week
	    col := 'hour_start';
	    secs := 3600;
	END IF;

	IF _direction = 'src' THEN
	    table_name := 'src_traffic';
	ELSIF _direction = 'dst' THEN
	    table_name := 'dst_traffic';
	ELSE
	    RETURN;
	END IF;

	query := '
          SELECT ' ||
            quote_ident(col) || ', sum(bytes) / ' || secs || '
          FROM ' || 
	    quote_ident(table_name) || '
          WHERE ' ||
            quote_ident(col) || ' BETWEEN ' || quote_literal(_start_time) || '::timestamp with time zone AND ' || 
                                               quote_literal(_end_time) || '::timestamp with time zone AND (';

	IF _network = 'all' THEN
	    i := 0;
	    FOR net IN
	      SELECT network FROM internal_net
	    LOOP
	      IF i > 0 THEN
	          query := query || ' OR ';
	      END IF;
	      query := query || 'addr <<= ' || quote_literal(net) || '::cidr';
	      i := i + 1;
	    END LOOP;
	ELSE
	    query := query || 'addr <<= ' || quote_literal(_network) || '::cidr';
	END IF;

	query := query || ') GROUP BY ' || quote_ident(col) || 
                          ' ORDER BY ' || quote_ident(col);

        EXECUTE 'CREATE TEMP TABLE bw_by_interval_results ON COMMIT DROP AS ' || query;
        SELECT count(*) INTO _count FROM bw_by_interval_results;
        OPEN _results FOR SELECT * FROM bw_by_interval_results;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_internal_servers function
--
-- retrieves servers located within internal networks (and within the specified timeframe)
-- --------------------
CREATE OR REPLACE FUNCTION get_internal_servers(varchar(18), timestamp with time zone, timestamp with time zone,
                                                varchar(6), varchar(4), OUT bigint, OUT refcursor) AS
$$
DECLARE
        _network    ALIAS FOR $1;
        _start_time ALIAS FOR $2;
        _end_time   ALIAS FOR $3;
        _column     ALIAS FOR $4;
        _order      ALIAS FOR $5;
        _count      ALIAS FOR $6;
        _results    ALIAS FOR $7;
	net         cidr;
        query       text;
        col         text;
	i	    int;
BEGIN
        query := '
            SELECT
              router, addr, port
            FROM
              src_traffic
            WHERE interval_start BETWEEN ' || quote_literal(_start_time) || '::timestamp with time zone AND ' || 
                                              quote_literal(_end_time) || '::timestamp with time zone AND
              port < 1024 AND (';

	IF _network = 'all' THEN
	    i := 0;
	    FOR net IN
	      SELECT network FROM internal_net
	    LOOP
	      IF i > 0 THEN
	          query := query || ' OR ';
	      END IF;
	      query := query || 'addr <<= ' || quote_literal(net) || '::cidr';
	      i := i + 1;
	    END LOOP;
	ELSE
	    query := query || 'addr <<= ' || quote_literal(_network) || '::cidr';
	END IF;

	query := query || ') GROUP BY router, addr, port ORDER BY ' || quote_ident(_column) || ' ' || _order;
	      
        EXECUTE 'CREATE TEMP TABLE get_internal_servers_results ON COMMIT DROP AS ' || query;
        SELECT count(*) INTO _count FROM get_internal_servers_results;
        OPEN _results FOR SELECT * FROM get_internal_servers_results;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_session_counts function
--
-- retrieves session counts (by protocol) within the specified timeframe
-- --------------------
CREATE OR REPLACE FUNCTION get_session_counts(varchar(18), timestamp with time zone, timestamp with time zone, 
                                              varchar(8), varchar(4), OUT bigint, OUT refcursor) AS
$$
DECLARE
        _network    ALIAS FOR $1;
        _start_time ALIAS FOR $2;
        _end_time   ALIAS FOR $3;
        _column     ALIAS FOR $4;
        _order      ALIAS FOR $5;
        _count      ALIAS FOR $6;
        _results    ALIAS FOR $7;
	net         cidr;
        query text;
	net_clause text;
	i int;
BEGIN
	IF _network = 'all' THEN
	    i := 0;
	    FOR net IN
	      SELECT network FROM internal_net
	    LOOP
	      IF i = 0 THEN
	          net_clause := '(';
	      ELSIF i > 0 THEN
	          net_clause := net_clause || ' OR ';
	      END IF;
	      net_clause := net_clause || 'src_addr <<= ' || quote_literal(net) || '::cidr OR dst_addr <<= ' || quote_literal(net) ||'::cidr';
	      i := i + 1;
	    END LOOP;
	    net_clause := net_clause || ')';
	ELSE
	    net_clause := '(src_addr <<= ' || quote_literal(_network) || '::cidr OR dst_addr <<= ' || quote_literal(_network) ||'::cidr)';
	END IF;

	query := 'SELECT * FROM ( SELECT * FROM (
          (SELECT
             protocol, count(*) as count
           FROM
	     (SELECT * FROM ((SELECT 6 as protocol FROM tcp_session
                              WHERE last_activity BETWEEN ' || quote_literal(_start_time) || '::timestamp with time zone AND ' || 
                                                             quote_literal(_end_time) || '::timestamp with time zone 
                                   AND ' || net_clause || ') 
	                    UNION ALL
                            (SELECT 6 as protocol FROM aged_tcp_session
                             WHERE last_activity BETWEEN ' || quote_literal(_start_time) || '::timestamp with time zone AND ' || 
                                                             quote_literal(_end_time) || '::timestamp with time zone
                                   AND ' || net_clause || ')) a ) b 
	   GROUP BY b.protocol )  
	  UNION ALL
          (SELECT
             protocol, count(*) as count
           FROM
	     (SELECT * FROM ((SELECT 17 as protocol FROM udp_session 
                              WHERE last_activity BETWEEN ' || quote_literal(_start_time) || '::timestamp with time zone AND ' || 
                                                               quote_literal(_end_time) || '::timestamp with time zone 
                                    AND ' || net_clause || ') 
	                    UNION ALL
                            (SELECT 17 as protocol FROM aged_udp_session 
                             WHERE last_activity BETWEEN ' || quote_literal(_start_time) || '::timestamp with time zone AND ' || 
                                                              quote_literal(_end_time) || '::timestamp with time zone 
                                   AND ' || net_clause || ')) a ) b
	   GROUP BY b.protocol ) 
	  UNION ALL
          (SELECT
             protocol, count(*) as count
           FROM
	     (SELECT * FROM ((SELECT prot as protocol FROM other_session 
                              WHERE last_activity BETWEEN ' || quote_literal(_start_time) || '::timestamp with time zone AND ' || 
                                                               quote_literal(_end_time) || '::timestamp with time zone 
                                    AND ' || net_clause || ') 
	                    UNION ALL
                            (SELECT prot as protocol FROM aged_other_session 
                             WHERE last_activity BETWEEN ' || quote_literal(_start_time) || '::timestamp with time zone AND ' || 
                                                              quote_literal(_end_time) || '::timestamp with time zone 
			          AND ' || net_clause || ')) a ) b
	   GROUP BY b.protocol ) ) c
           ORDER BY c.' || quote_ident(_column) || ' ' || _order || ' ) d';

        EXECUTE 'CREATE TEMP TABLE get_session_counts_results ON COMMIT DROP AS ' || query;
        SELECT count(*) INTO _count FROM get_session_counts_results;
        OPEN _results FOR SELECT * FROM get_session_counts_results;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_traffic
--
-- retrieves the requested rows from the src_traffic or dst_traffic table
-- --------------------
CREATE OR REPLACE FUNCTION get_traffic(varchar(18), varchar(16), varchar(3), 
                                       integer[][], integer[], integer[][], integer[], 
                                       integer[][], integer[], integer[][], integer[], 
                                       timestamp with time zone, timestamp with time zone, 
				       varchar(16), varchar(4),
				       OUT bigint, OUT refcursor) AS
$$
DECLARE
	_network     ALIAS FOR $1;
	_router      ALIAS FOR $2;
	_dir         ALIAS FOR $3;
	_iprot_range ALIAS FOR $4;
	_iprot       ALIAS FOR $5;
	_eprot_range ALIAS FOR $6;
	_eprot       ALIAS FOR $7;
	_iport_range ALIAS FOR $8;
	_iport       ALIAS FOR $9;
	_eport_range ALIAS FOR $10;
	_eport       ALIAS FOR $11;
        _start_time  ALIAS FOR $12;
        _end_time    ALIAS FOR $13;
        _column      ALIAS FOR $14;
        _order       ALIAS FOR $15;
        _count       ALIAS FOR $16;
        _results     ALIAS FOR $17;
	table_name   text;
        time_clause  text;
        net_clause   text := ' ';
        router_clause text := ' ';
        prot_clause  text := ' ';
        port_clause  text := ' ';
        query text;
BEGIN
	_count := 0; -- set default count in case function returns without a result set

	-- check the direction of traffic to determine the table to query
	IF _dir = 'src' THEN
	  table_name = 'src_traffic';
	ELSIF _dir = 'dst' THEN
	  table_name = 'dst_traffic';
	ELSE
	  RETURN;
	END IF;

	-- based on the input parameters, assemble the where clauses used in the query below
	SELECT * INTO time_clause   FROM get_time_clause('interval_start', _start_time, _end_time);
	SELECT * INTO net_clause    FROM get_net_clause(0, _network);
	SELECT * INTO router_clause FROM get_router_clause(_router);
	SELECT * INTO prot_clause   FROM get_prot_clause(_iprot_range, _iprot, _eprot_range, _eprot);
	SELECT * INTO port_clause   FROM get_port_clause(0, _iport_range, _iport, _eport_range, _eport);

        query := 'SELECT interval_start, router, addr, prot, port, pkts, bytes
	          FROM ' || quote_ident(table_name) || '
	          WHERE ' || time_clause || net_clause || router_clause || prot_clause || port_clause || '
                  ORDER BY ' || quote_ident(_column) || ' ' || _order;

        EXECUTE 'CREATE TEMP TABLE traffic_results ON COMMIT DROP AS ' || query;

        SELECT count(*) INTO _count FROM traffic_results;
        OPEN _results FOR SELECT * FROM traffic_results;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_sessions function
--
-- retrieves the requested rows from the session tables
-- --------------------
CREATE OR REPLACE FUNCTION get_sessions(varchar(18), varchar(16), 
                                        integer[][], integer[], integer[][], integer[], 
                                        integer[][], integer[], integer[][], integer[], 
                                        real, real, timestamp with time zone, timestamp with time zone, 
					varchar(16), varchar(4), OUT bigint, OUT refcursor) AS
$$
DECLARE
	_network     ALIAS FOR $1;
	_router      ALIAS FOR $2;
	_iprot_range ALIAS FOR $3;
	_iprot       ALIAS FOR $4;
	_eprot_range ALIAS FOR $5;
	_eprot       ALIAS FOR $6;
	_iport_range ALIAS FOR $7;
	_iport       ALIAS FOR $8;
	_eport_range ALIAS FOR $9;
	_eport       ALIAS FOR $10;
	_min_ratio   ALIAS FOR $11;
	_max_ratio   ALIAS FOR $12;
        _start_time  ALIAS FOR $13;
        _end_time    ALIAS FOR $14;
        _column      ALIAS FOR $15;
        _order       ALIAS FOR $16;
        _count       ALIAS FOR $17;
        _results     ALIAS FOR $18;
        time_clause  text;
        net_clause   text := ' ';
        router_clause text := ' ';
        prot_clause  text := ' ';
        port_clause  text := ' ';
        div_clause   text;
	i integer := 1;
        query text;
BEGIN
	_count := 0; -- set default count in case function returns without a result set

	-- sanity check the min and max ratios
	IF _min_ratio <= 0 THEN
	  RETURN;
	ELSIF _max_ratio <= _min_ratio THEN
	  RETURN;
	END IF;

	-- based on the input parameters, assemble the where clauses used in the query below
	SELECT * INTO time_clause   FROM get_time_clause('interval_start', _start_time, _end_time);
	SELECT * INTO net_clause    FROM get_net_clause(1, _network);
	SELECT * INTO router_clause FROM get_router_clause(_router);
	SELECT * INTO prot_clause   FROM get_prot_clause(_iprot_range, _iprot, _eprot_range, _eprot);
	SELECT * INTO port_clause   FROM get_port_clause(1, _iport_range, _iport, _eport_range, _eport);
	SELECT * INTO div_clause    FROM get_div_clause();

	-- Find the sessions that match the input criteria
        query := 'SELECT * FROM 
		  ((SELECT session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, 6 as prot, 
                           in_bytes, out_bytes, (in_bytes::real / out_bytes::real) as ratio
	            FROM tcp_session
	            WHERE ' || time_clause || net_clause || router_clause || port_clause || div_clause || ')
		   UNION ALL	
		   (SELECT session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, 6 as prot, 
                           in_bytes, out_bytes, (in_bytes::real / out_bytes::real) as ratio
	            FROM aged_tcp_session
	            WHERE ' || time_clause || net_clause || router_clause || port_clause || div_clause || ')) a';
	IF prot_clause != ' ' THEN
	  query := query || ' WHERE 1=1 ' || prot_clause;
	END IF;
        EXECUTE 'CREATE TEMP TABLE temp_tcp_sessions ON COMMIT DROP AS ' || query;

        query := 'SELECT * FROM 
		  ((SELECT session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, 17 as prot, 
                           in_bytes, out_bytes, (in_bytes::real / out_bytes::real) as ratio
	            FROM udp_session
	            WHERE ' || time_clause || net_clause || router_clause || port_clause || div_clause || ')
		   UNION ALL	
		   (SELECT session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, 17 as prot, 
                           in_bytes, out_bytes, (in_bytes::real / out_bytes::real) as ratio
	            FROM aged_udp_session
	            WHERE ' || time_clause || net_clause || router_clause || port_clause || div_clause || ')) a';
	IF prot_clause != ' ' THEN
	  query := query || ' WHERE 1=1 ' || prot_clause;
	END IF;
        EXECUTE 'CREATE TEMP TABLE temp_udp_sessions ON COMMIT DROP AS ' || query;

        query := 'SELECT * FROM 
		  ((SELECT session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, prot, 
                           in_bytes, out_bytes, (in_bytes::real / out_bytes::real) as ratio
	            FROM other_session
	            WHERE ' || time_clause || net_clause || router_clause || prot_clause || port_clause || div_clause || ')
		   UNION ALL	
		   (SELECT session_start, last_activity, router, src_addr, dst_addr, src_port, dst_port, prot, 
                           in_bytes, out_bytes, (in_bytes::real / out_bytes::real) as ratio
	            FROM aged_other_session
	            WHERE ' || time_clause || net_clause || router_clause || prot_clause || port_clause || div_clause || ')) a';
        EXECUTE 'CREATE TEMP TABLE temp_other_sessions ON COMMIT DROP AS ' || query;

	query := 'SELECT * FROM (SELECT * FROM temp_tcp_sessions 
                                 UNION ALL 
                                 SELECT * FROM temp_udp_sessions 
                                 UNION ALL 
                                 SELECT * FROM temp_other_sessions) a 
		  WHERE ratio BETWEEN ' || quote_literal(_min_ratio) || ' AND ' || quote_literal(_max_ratio) || ' 
                  ORDER BY ' || quote_ident(_column) || ' ' || _order;
        EXECUTE 'CREATE TEMP TABLE sessions_results ON COMMIT DROP AS ' || query;

        SELECT count(*) INTO _count FROM sessions_results;
        OPEN _results FOR SELECT * FROM sessions_results;
END;
$$
LANGUAGE plpgsql;

-- --------------------
-- get_flows function
--
-- retrieves the requested rows from the flow table
-- --------------------
CREATE OR REPLACE FUNCTION get_flows(varchar(18), varchar(18), varchar(16), 
                                     integer[][], integer[], integer[][], integer[], 
                                     integer[][], integer[], integer[][], integer[], 
                                     timestamp with time zone, timestamp with time zone, 
		      	             varchar(12), varchar(4), OUT bigint, OUT refcursor) AS
$$
DECLARE
	_src_network ALIAS FOR $1;
	_dst_network ALIAS FOR $2;
	_router      ALIAS FOR $3;
	_iprot_range ALIAS FOR $4;
	_iprot       ALIAS FOR $5;
	_eprot_range ALIAS FOR $6;
	_eprot       ALIAS FOR $7;
	_iport_range ALIAS FOR $8;
	_iport       ALIAS FOR $9;
	_eport_range ALIAS FOR $10;
	_eport       ALIAS FOR $11;
        _start_time  ALIAS FOR $12;
        _end_time    ALIAS FOR $13;
        _column      ALIAS FOR $14;
        _order       ALIAS FOR $15;
        _count       ALIAS FOR $16;
        _results     ALIAS FOR $17;
        time_clause  text;
        net_clause   text := ' ';
        router_clause text := ' ';
        prot_clause  text := ' ';
        port_clause  text := ' ';
        query text;
BEGIN
	-- based on the input parameters, assemble the where clauses used in the query below
	SELECT * INTO time_clause   FROM get_time_clause('start_time', _start_time, _end_time);
	SELECT * INTO net_clause    FROM get_net_clause(_src_network, _dst_network);
	SELECT * INTO router_clause FROM get_router_clause(_router);
	SELECT * INTO prot_clause   FROM get_prot_clause(_iprot_range, _iprot, _eprot_range, _eprot);
	SELECT * INTO port_clause   FROM get_port_clause(1, _iport_range, _iport, _eport_range, _eport);

	-- Find the flows that match the input criteria
        query := 'SELECT * FROM flow
	          WHERE ' || time_clause || net_clause || router_clause || prot_clause || port_clause || '
                  ORDER BY ' || quote_ident(_column) || ' ' || _order;
        EXECUTE 'CREATE TEMP TABLE flows_results ON COMMIT DROP AS ' || query;

        SELECT count(*) INTO _count FROM flows_results;
        OPEN _results FOR SELECT * FROM flows_results;
END;
$$
LANGUAGE plpgsql;

-- ---------------------
-- get_internal_nets
--
-- retrieves rows from the internal_net table
-- --------------------
CREATE OR REPLACE FUNCTION get_internal_nets(varchar(4), OUT bigint, OUT refcursor) AS
$$
DECLARE
        _order       ALIAS FOR $1;
        _count       ALIAS FOR $2;
        _results     ALIAS FOR $3;
	query text;
BEGIN
	query := 'SELECT network FROM internal_net ORDER by network ' || _order;
        EXECUTE 'CREATE TEMP TABLE get_internal_nets_results ON COMMIT DROP AS ' || query;

        SELECT count(*) INTO _count FROM get_internal_nets_results;
        OPEN _results FOR SELECT * FROM get_internal_nets_results;
END;
$$
LANGUAGE plpgsql;

-- ---------------------
-- add_internal_net
--
-- adds a network or IP (in CIDR notation) to the internal_net table
-- --------------------
CREATE OR REPLACE FUNCTION add_internal_net(cidr) RETURNS void AS
$$
DECLARE
        _network     ALIAS FOR $1;
BEGIN
	UPDATE
	  internal_net
	SET
	  network = network
	WHERE
	  network = _network::cidr;
	IF NOT FOUND THEN
          INSERT INTO
	    internal_net(network) 
	  VALUES
	    (_network::cidr);
	END IF;

	RETURN;
END;
$$
LANGUAGE plpgsql;

-- ---------------------
-- remove_internal_net
--
-- removes a network or IP (in CIDR notation) from the internal_net table
-- --------------------
CREATE OR REPLACE FUNCTION remove_internal_net(cidr) RETURNS void AS
$$
DECLARE
        _network     ALIAS FOR $1;
	_count int;
BEGIN
	SELECT count(*) FROM internal_net INTO _count;

	IF _count > 1 THEN
          DELETE FROM
	    internal_net 
	   WHERE
	    network = _network::cidr;
	END IF;

	RETURN;
END;
$$
LANGUAGE plpgsql;
