#ifndef NETFLOW_H_GUARD
#define NETFLOW_H_GUARD

/* #include "database.h" */

#define NFV5_HEADER_LEN 24
#define NFV5_RECORD_LEN 48

#pragma pack(push, 1)

typedef struct
{
    PRUint16 version;       /* netflow export version number 0x0005 */
    PRUint16 count;         /* number of flows exported in this datagram */
    PRUint32 sysUptime;     /* current time in msec since the router booted */
    PRUint32 unix_secs;     /* Seconds since 0000 UTC 1970 */
    PRUint32 unix_nsecs;    /* Residual nanoseconds since 0000 UTC 1970 */
    PRUint32 flow_sequence; /* Sequence counter of total flow seen */

    /* NOTE: the Cisco documentation at
     * http://www.cisco.com/en/US/docs/ios/12_4/netflow/configuration/guide/onf_bcf.html,
     * is confusing about whether the following reserved field actually exists; captured
     * datagrams indicate it does not.
    PRUint32 reserved; */

    PRUint8  engine_type;   /* Type of flow switching engine: 0 for RP, 1 for VIP/LC */
    PRUint8  engine_ID;     /* VIP or LC slot number of the flow switching engine */
    PRUint16 pad;
}
nfv5_header_t;

typedef nfv5_header_t host_header_t; /* converted to host byte order */

typedef struct
{
    PRUint32 srcaddr;   /* Source IP address */
    PRUint32 dstaddr;   /* Destination IP address */
    PRUint32 nexthop;   /* Next hop router's IP address */
    PRUint16 input;     /* Ingress interface SNMP ifindex */
    PRUint16 output;    /* Egress interface SNMP ifindex */
    PRUint32 dPkts;     /* Packets in the flow */
    PRUint32 dOctets;   /* Octets (bytes) in the flow */
    PRUint32 first;     /* SysUptime at start of the flow */
    PRUint32 last;      /* SysUptime at the time the last packet of the flow was received */
    PRUint16 srcport;   /* Layer 4 source port number or equivalent */
    PRUint16 dstport;   /* Layer 4 destination port number or equivalent */
    PRUint8  pad1;      /* Unused (zero) byte */
    PRUint8  tcp_flags; /* Cumulative OR of TCP flags */
    PRUint8  prot;      /* Layer 4 protocol (for example, 6=TCP, 17=UDP) */
    PRUint8  tos;       /* IP type-of-service byte */
    PRUint16 src_as;    /* Autonomous system number of the source, either origin or peer */
    PRUint16 dst_as;    /* Autonomous system number of the destination, either origin or peer */
    PRUint8  src_mask;  /* Source address prefix mask bits */
    PRUint8  dst_mask;  /* Destination address prefix mask bits */
    PRUint16 pad2;      /* PAD 2 is unused (zero) bytes */
}
nfv5_record_t;

typedef nfv5_record_t host_record_t; /* converted to host byte order */

#pragma pack(pop)

/* process netflow v5 datagram */
PR_EXTERN(void) process_nfv5_datagram(db_data_t*, const char*, nfv5_header_t*, PRInt32);

/* convert netflow v5 header to host byte order */
PR_EXTERN(void) nfv5_header_ntoh(host_header_t*, const nfv5_header_t*);

/* convert netflow v5 record to host byte order */
PR_EXTERN(void) nfv5_record_ntoh(host_record_t*, const nfv5_record_t*);

#endif  /* NETFLOW_H_GUARD */
