#ifndef PFCAPD_H_GUARD
#define PFCAPD_H_GUARD

#include <nspr.h>
#include <plgetopt.h>
#include <prlog.h>
#include <prerror.h>
#include <prnetdb.h>
#include "config.h"
#include "globals.h"
#include "conf.h"
#include "prerror.h"
#include "whitelist.h"
#include "netflow.h"
#include "database.h"
#include "threads.h"

#ifndef XP_UNIX
#error "pfcapd currently only builds and runs on Unix"
#endif

/* documentation macro, for converting between nano- and micro-seconds */
#define PR_NSEC_PER_USEC PR_USEC_PER_MSEC

#ifdef HAVE_DAEMON
#include <unistd.h>
#define DAEMONIZE(nochdir, noclose) daemon(nochdir, noclose)
#else
#define DAEMONIZE(nochdir, noclose) PR_LOG(globals.dlogger, PR_LOG_DEBUG, ("daemon mode not available"))
#endif

#define UDP_BUFFER_SIZE 65535   /* maximum UDP datagram size */

#define PR_LOG_TIMER1 PR_LOG_ERROR /* datagram-level timers */
#define PR_LOG_TIMER2 PR_LOG_WARN  /* datagram- and database-level timers */

#ifdef DEBUG
#define INCR_DGCOUNT(result) PRInt32 result = PR_AtomicIncrement(&globals.dgCount)
#define LOG_DEBUG(m, l, a) PR_LOG(m, l, a)
#define LOG_TICK(id) PR_LOG(globals.dlogger, PR_LOG_TIMER1, ("tick=%u " #id, PR_IntervalNow()))
#define LOG_MSEC(id) PR_LOG(globals.dlogger, PR_LOG_TIMER1, ("msec=%u " #id, PR_IntervalToMilliseconds(PR_IntervalNow())))
#define START_TIMER1(epoch) epoch = PR_IntervalNow()
#define STOP_TIMER1(epoch) PR_LOG(globals.dlogger, PR_LOG_TIMER1, (#epoch " timer stopped at %u msec", PR_IntervalToMilliseconds(PR_IntervalNow() - epoch)))
#define STOP_TIMER1_LOG_ROUTER(router,epoch) PR_LOG(globals.dlogger, PR_LOG_TIMER1, ("router=%s " #epoch " timer stopped at %u msec", router, PR_IntervalToMilliseconds(PR_IntervalNow() - epoch)))
#define START_TIMER2(epoch) epoch = PR_IntervalNow()
#define STOP_TIMER2(epoch) PR_LOG(globals.dlogger, PR_LOG_TIMER2, (#epoch " timer stopped at %u msec", PR_IntervalToMilliseconds(PR_IntervalNow() - epoch)))
#else  /* !DEBUG */
#define INCR_DGCOUNT(result)    /* do nothing */
#define LOG_DEBUG(m, l, a)      /* do nada */
#define LOG_TICK(id)            /* do zip */
#define LOG_MSEC(id)            /* do zilch */
#define START_TIMER1(epoch)     /* do nil */
#define STOP_TIMER1(epoch)      /* do null */
#define STOP_TIMER1_LOG_ROUTER(router,epoch)
#define START_TIMER2(epoch)     /* do nil */
#define STOP_TIMER2(epoch)      /* do null */
#endif  /* DEBUG */

#endif  /* PFCAPD_H_GUARD */
