import turbogears as tg
from turbogears import controllers, expose, flash, validate, error_handler, config
from turbogears import identity, redirect
from cherrypy import request, response
from turbogears.widgets.base import WidgetsList
from turbogears.widgets import TextField, CheckBox, RadioButtonList, RemoteForm, Widget, AjaxGrid, TableForm
from turbogears.widgets import CalendarDateTimePicker, CompoundFormField, SingleSelectField
from turbogears import validators
from turbojson.jsonify import jsonify
import formencode
from psychefe.widgets.turboflotr import TurboFlotr
import datetime
import re, time
from math import *
from decimal import *
import customValidators, storedProcedures, grid_options
from TGExtJS.widgets import PagingJSONGrid
import operator
import random
import logging
import csv
from cherrypy.lib.cptools import serveFile
import os
import glob

log = logging.basicConfig(filename='debug.log', level=logging.DEBUG)
working_dir = os.getcwd() + "/"

@jsonify.when("isinstance(obj, Widget)")
def jsonify_widget(obj):
	return None

#----------------------------------------------
# Colors used for the top talkers bar graph
#----------------------------------------------
ttColors = { 0:'blue', 
	   1:'green', 
	   2:'orange', 
	   3:'yellow', 
	   4:'white',
	   5:'red',
	   6:'purple',
	   7:'pink',
	   8:'brown', 
	   9:'maroon' }

#----------------------------------------------
# Javascript helper functions passed to TurboFlotr
#----------------------------------------------
timeSeries = """
		function timestampHMTickFormatter(inputTimeStamp) {
			mydate = new Date();
			mydate.setTime(inputTimeStamp*1000);
			hours = mydate.getHours(); // -1 to get real numbers
			minutes = mydate.getMinutes();
			if (minutes < 10) { // adding zero
				minutes = "0"+minutes;
			}
			output = hours + ":" + minutes;
			return output;
		}

		function timestampDHMTickFormatter(inputTimeStamp) {
			days = new Array();
			days[0] = "Sun";
			days[1] = "Mon";
			days[2] = "Tue";
			days[3] = "Wed";
			days[4] = "Thu";
			days[5] = "Fri";
			days[6] = "Sat";
			mydate = new Date();
			mydate.setTime(inputTimeStamp*1000);
			day = mydate.getDay();
			hours = mydate.getHours(); // -1 to get real numbers
			minutes = mydate.getMinutes();
			if (hours < 10) { // adding zero
				hours = "0"+hours;
			}
			if (minutes < 10) { // adding zero
				minutes = "0"+minutes;
			}
			output = days[day] + "   " + hours + ":" + minutes;
			return output;
		}

		function bwTickFormatter(inBytes) { return (inBytes); }
		function dateBWTrackFormatter(obj) { return timestampHMTickFormatter(obj.x) + ', ' + obj.y ;}
		typeGraph = 'timeseries';
"""

histogram = """
		function ratioTickFormatter(n) {
			return Math.pow(10,n).toFixed(2);
		}

		function histoTrackFormatter(obj) { return ratioTickFormatter(obj.x) + ', ' + obj.y ;}
		typeGraph = 'histogram';
"""

#----------------------------------------------
# Flotr and Grid objects used for display
#----------------------------------------------

homeFlotr = TurboFlotr(refresh_url="/UpdateHomeGraph", form_id="homeForm",
		       id="lineGraph", height='300px', width='600px', helperFunc=timeSeries)
topTalkersFlotr = TurboFlotr(refresh_url="/UpdateTopTalkers", form_id="ttForm", 
                             id="barGraph", height='400px', width='800px', helperFunc=timeSeries)
histogramFlotr = TurboFlotr(refresh_url="/UpdateHistogram", form_id="histForm", 
                            id="barGraph", height='400px', width='800px', marginBottom='5em', helperFunc=histogram)
topTalkersGrid = AjaxGrid(refresh_url="/getColumnInfo", refresh_text="")
paging_grid = PagingJSONGrid()

#----------------------------------------------
# Helper function that grabs the internal networks from the DB
#----------------------------------------------
def getInternalNets():
	kw = {}
	kw['order'] = "'ASC'"
	dbCursor = storedProcedures.CursorControl(kw, 'get_internal_nets')
	total = int(dbCursor.count())
	rows = dbCursor.getResults(0, total)
	options = []
	options.append("all internal networks")
	for res in rows:
	  options.append(str(res[0]))
	return options

#----------------------------------------------
# Generic Compound Forms
#----------------------------------------------
class CompTimeArea(CompoundFormField):
	member_widgets = ["start_time", "end_time"]

	template = """
	      <div xmlns:py="http://purl.org/kid/ns#">
		      <br/>${display_field_for(start_time)}
		      <br/>${display_field_for(end_time)}
		      <br/><hr/>
	      </div>
	"""
	start_time = CalendarDateTimePicker(name="start_time", label="Start Time", skin="calendar-blue",
					    validator=validators.All(validators.DateTimeConverter(), validators.NotEmpty()),
					    button_text="Select Start Time", default = datetime.datetime.now() - datetime.timedelta(hours = 4))
	end_time = CalendarDateTimePicker(name="end_time", label="End Time", skin="calendar-blue",
				          validator=validators.All(validators.DateTimeConverter(), validators.NotEmpty()),
					  button_text="Select End Time", default = datetime.datetime.now())

	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompTimeArea, self).__init__(*args, **kw)

class CompNetworkArea(CompoundFormField):
	member_widgets = ["network", "router"]

	template = """
	<div xmlns:py="http://purl.org/kid/ns#">
		<br/>Name or IP of Exporter: ${display_field_for(router)}
		<br/><br/> Network of Interest: ${display_field_for(network)}
		<br/><br/>
	</div>
	"""

	#at some point tfield_params will have target neworks prepopulated
	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompNetworkArea, self).__init__(*args, **kw)
		#TODO: custom validator to strip off netmask, if one is present
		self.router = TextField(name="router", label="Netflow Exporter")
		self.network = TextField(name="network", label="Target Network", validator=validators.CIDR())

class CompProtsPortsArea(CompoundFormField):
	member_widgets = ["portIncList", "portExcList", "protIncList", "protExcList", "protocolsPortsRadio"]

	template = """
	<div xmlns:py="http://purl.org/kid/ns#">
		<hr/><br/>Note: inputs can be single elements and/or ranges (ex: 1,17-20)
		<br/><br/>Ports to include: ${display_field_for(portIncList)}
		<br/><br/>Ports to exclude: ${display_field_for(portExcList)}
		<br/><br/>Protocols to include: ${display_field_for(protIncList)}
		<br/><br/>Protocols to exclude: ${display_field_for(protExcList)}
		<br/><br/>
	</div>
	"""

	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompProtsPortsArea, self).__init__(*args, **kw)
		self.portIncList = TextField(name="portIncList")	#TODO: add Validator
		self.portExcList = TextField(name="portExcList")	#TODO: add Validator
		self.protIncList = TextField(name="protIncList")	#TODO: add Validator
		self.protExcList = TextField(name="protExcList")	#TODO: add Validator

class CompDirectionArea(CompoundFormField):
	member_widgets = ["direction"]

	template = """
	<div xmlns:py="http://purl.org/kid/ns#">
                Direction of traffic: ${display_field_for(direction)}
		<br/>
	</div>
	"""

	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompDirectionArea, self).__init__(*args, **kw)
		self.direction = RadioButtonList(name="direction", label="", options=[(1, "Source"), (2, "Destination")], default=1)

class TimeSchema(validators.Schema):
	'''verify start time is before end time'''
	chained_validators = [customValidators.DatesValidator()]

class TimeSchemaMaxInt(validators.Schema):
	'''verify start time is before end time'''
	chained_validators = [customValidators.DatesValidator(maxInterval=datetime.timedelta(days=2))]

#----------------------------------------------
# Home Compound Forms
#----------------------------------------------
class CompHomeNetworkArea(CompoundFormField):
 	member_widgets = ["internal_net"]

	template = """
		<div xmlns:py="http://purl.org/kid/ns#">
		    Show data for: ${display_field_for(internal_net)}
		    <br/>
		    <br/>
		</div>
	"""

	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompHomeNetworkArea, self).__init__(*args, **kw)
		self.internal_net = SingleSelectField("internal_net", options=getInternalNets, default=1)
		
class CompHomeTimeArea(CompoundFormField):
 	member_widgets = ["time_interval"]

	template = """
		<div xmlns:py="http://purl.org/kid/ns#">
		    Show data for the past: ${display_field_for(time_interval)}
		    <br/>
		    <br/>
		</div>
	"""

	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompHomeTimeArea, self).__init__(*args, **kw)
		self.time_interval = SingleSelectField("time_interval", options=[#(1, "hour"),
									         (2, "day"),
									         (3, "week")],
			  						         default=2)
		
#----------------------------------------------
# Top Talkers Compound Forms
#----------------------------------------------
class CompTopTalkersCountArea(CompoundFormField):
 	member_widgets = ["num_talkers", "talker_type"]

	template = """
		<div xmlns:py="http://purl.org/kid/ns#">
		    Show data for the top: ${display_field_for(num_talkers)} ${display_field_for(talker_type)}
		    <br/>
		</div>
	"""

	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompTopTalkersCountArea, self).__init__(*args, **kw)
		self.num_talkers = SingleSelectField("num_talkers", options=[(1, "1"),
									     (2, "2"),
									     (3, "3"),
									     (4, "4"),
									     (5, "5"),
									     (6, "6"),
									     (7, "7"),
									     (8, "8"),
									     (9, "9")],
			  						     default=1)
		self.talker_type = SingleSelectField("talker_type", options=[(1, "port(s)"),
									 (2, "host(s)")],
			  						 default=1)
		

#----------------------------------------------
# Histogram Compound Forms
#----------------------------------------------
class CompHistogramGraphOptionsArea(CompoundFormField):
	member_widgets = ["min_ratio", "max_ratio"]

	template = """
	<div xmlns:py="http://purl.org/kid/ns#">
		<hr/><br/>
		Min Ratio: ${display_field_for(min_ratio)}<br/>
		<br/>
		Max Ratio: ${display_field_for(max_ratio)}
		<br/>
		<br/>
	</div>
	"""

	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompHistogramGraphOptionsArea, self).__init__(*args, **kw)
		self.min_ratio = TextField( name="min_ratio", validator=validators.Number(), default=0.1)
		self.max_ratio = TextField( name="max_ratio",	validator=validators.Number(), default=10.0)

#----------------------------------------------
# Raw Flows Compound Forms
#----------------------------------------------
class CompRawFlowsNetworkArea(CompoundFormField):
	member_widgets = ["src_network", "dst_network", "router"]

	template = """
	<div xmlns:py="http://purl.org/kid/ns#">
		<br/>Name or IP of Exporter: ${display_field_for(router)}
		<br/><br/> Src Network of Interest: ${display_field_for(src_network)}
		<br/><br/> Dst Network of Interest: ${display_field_for(dst_network)}
		<br/><br/>
	</div>
	"""

	#at some point tfield_params will have target neworks prepopulated
	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompRawFlowsNetworkArea, self).__init__(*args, **kw)
		#TODO: custom validator to strip off netmask, if one is present
		self.router = TextField(name="router", label="Netflow Exporter")
		self.src_network = TextField(name="src_network", label="Target Src Network", validator=validators.CIDR())
		self.dst_network = TextField(name="dst_network", label="Target Dst Network", validator=validators.CIDR())

#----------------------------------------------
# Management Compound Forms
#----------------------------------------------
class CompMgmtActionArea(CompoundFormField):
	member_widgets = ["action"]

	template = """
	<div xmlns:py="http://purl.org/kid/ns#">
                ${display_field_for(action)}
		<br/>
	</div>
	"""

	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompMgmtActionArea, self).__init__(*args, **kw)
		self.action = RadioButtonList(name="action", label="", options=[(1, "Show Networks"), (2, "Add Network"), (3, "Remove Network")], default=1)

class CompMgmtNetworkArea(CompoundFormField):
	member_widgets = ["mgmt_network"]

	template = """
	<div xmlns:py="http://purl.org/kid/ns#">
		Network to Add or Remove: ${display_field_for(mgmt_network)}
		<br/><br/>Note: at least one internal network must always exist
		<br/><br/>
	</div>
	"""

	#at some point tfield_params will have target neworks prepopulated
	def __init__(self, tfield_params={}, sel_params={}, *args, **kw):
		super(CompMgmtNetworkArea, self).__init__(*args, **kw)
		#TODO: custom validator to strip off netmask, if one is present
		self.mgmt_network = TextField(name="mgmt_network", label="Target Network", validator=validators.CIDR())

#----------------------------------------------
# Put our forms together
#----------------------------------------------
class histogramFormFields(WidgetsList):
	timeArea = CompTimeArea(label="", validator=TimeSchema())
	targetNetworkArea = CompNetworkArea(label="")
	protocolsPortsArea = CompProtsPortsArea(label="")
	graphOptionsArea = CompHistogramGraphOptionsArea(label="")

class topTalkersFormFields(WidgetsList):
	countArea = CompTopTalkersCountArea(label="")
	timeArea = CompTimeArea(label="", validator=TimeSchemaMaxInt())
	targetNetworkArea = CompNetworkArea(label="")
	directionArea = CompDirectionArea(label="")
	protocolsPortsArea = CompProtsPortsArea(label="")

class rawFlowsFormFields(WidgetsList):
	timeArea = CompTimeArea(label="", validator=TimeSchema())
	targetNetworkArea = CompRawFlowsNetworkArea(label="")
	protocolsPortsArea = CompProtsPortsArea(label="")

class homeFormFields(WidgetsList):
	targetNetworkArea = CompHomeNetworkArea(label="")
	directionArea = CompDirectionArea(label="")
	timeArea = CompHomeTimeArea(label="")

class mgmtFormFields(WidgetsList):
	actionArea = CompMgmtActionArea(label="")
	targetNetworkArea = CompMgmtNetworkArea(label="")

#----------------------------------------------
# Root Class
#----------------------------------------------
class Root(controllers.RootController):

	#first clean up any CSV files left in the working directory
	for filename in glob.glob(working_dir + '*.csv'):
        	absPath = os.path.abspath(filename)
		os.remove(absPath)

	homeForm = RemoteForm(name="form", fields=homeFormFields(), submit_text="Submit")
	topTalkersForm = RemoteForm(name="form", fields=topTalkersFormFields(), validator=TimeSchemaMaxInt(), submit_text="Submit")
	histogramForm = RemoteForm(name="form", fields=histogramFormFields(), validator=TimeSchema(), submit_text="Submit")
	rawFlowsForm = RemoteForm(name="form", fields=rawFlowsFormFields(), validator=TimeSchema(), submit_text="Submit")
	mgmtForm = RemoteForm(name="form", fields=mgmtFormFields(), submit_text="Submit")

	#TODO: change to map session id's to parameters
	mgmtSessParams = {}
	rawSessParams = {}
	homeSessParams = {}
	histSessParams = {}
	ttSessParams = {}

	#stores our session data for the flow grids
	mgmtGridSessions = {}
	rawGridSessions = {}
	histGridSessions = {}
	ttGridSessions = {}
	homeGridSessions = {}

	#TODO: move these into the storedProcedures.py class
	dispatcher = storedProcedures.SPDispatcher(config.get('psyche.dburi'))

	#----------------------------------------------
	# Identity management stuff
	#----------------------------------------------
	@expose(template="psychefe.templates.login")
	def login(self, forward_url=None, previous_url=None, *args, **kw):

		if not identity.current.anonymous and identity.was_login_attempted() \
				and not identity.get_identity_errors():
			redirect(tg.url(forward_url or previous_url or '/', kw))

		forward_url = None
		previous_url = request.path

		if identity.was_login_attempted():
			msg = _("The credentials you supplied were incorrect or "
				   "you are not authorized to access Psyche.")
		elif identity.get_identity_errors():
			msg = _("You must provide your credentials before accessing Psyche.")
		else:
			msg = _("Please log in.")
			forward_url = request.headers.get("Referer", "/")

		response.status = 403
		return dict(message=msg, previous_url=previous_url, logging_in=True,
			original_parameters=request.params, forward_url=forward_url)

	@expose()
	def logout(self):
		identity.current.logout()
		redirect("/")
	
	@tg.expose("json")
	def raw_grid_settings(self):
		return grid_options.raw_flow_list

	@tg.expose("json")
	def hist_grid_settings(self):
		return grid_options.hist_flow_list

	@tg.expose("json")
	def tt_grid_settings(self):
		return grid_options.tt_flow_list

	@tg.expose("json")
	def home_server_grid_settings(self):
		return grid_options.home_server_list

	@tg.expose("json")
	def home_session_grid_settings(self):
		return grid_options.home_session_list

	@tg.expose("json")
	def mgmt_server_grid_settings(self):
		return grid_options.mgmt_server_list

	#----------------------------------------------
	# Our Pages
	# --index (home)
	# --topTalkers
	# --histogram
	# --rawFlows
	# --management
	#----------------------------------------------
	@expose(template="psychefe.templates.welcome")
	@identity.require(identity.in_any_group("admin", "analyst"))
	def index(self, tg_errors = None):
		self.removePageSession('mgmtSessionId', self.mgmtGridSessions)
		self.removePageSession('homeSessionId', self.homeGridSessions)
		self.removePageSession('rawSessionId', self.rawGridSessions)
		self.removePageSession('ttSessionId', self.ttGridSessions)
		self.removePageSession('histSessionId', self.histGridSessions)
		return dict(flotr=homeFlotr, homeForm=self.homeForm,
			    rows=[], headers=[],
			    paging_grid=paging_grid,
			    server_data_url = '/home_server_grid_data',
			    session_data_url = '/home_session_grid_data',
			    server_property_url='/home_server_grid_settings',
			    session_property_url='/home_session_grid_settings')

	@expose(template="psychefe.templates.topTalkers")
	@identity.require(identity.in_any_group("admin", "analyst"))
	def topTalkers(self, tg_errors = None):
		self.removePageSession('mgmtSessionId', self.mgmtGridSessions)
		self.removePageSession('homeSessionId', self.homeGridSessions)
		self.removePageSession('rawSessionId', self.rawGridSessions)
		self.removePageSession('ttSessionId', self.ttGridSessions)
		self.removePageSession('histSessionId', self.histGridSessions)
		return dict(flotr=topTalkersFlotr, ttForm=self.topTalkersForm,
			    topTalkersGrid=topTalkersGrid,
			    rows=[], headers=[],
			    paging_grid=paging_grid,
			    data_url = '/tt_grid_data',
			    property_url='/tt_grid_settings')

	@expose(template="psychefe.templates.histogram")
	@identity.require(identity.in_any_group("admin", "analyst"))
	def histogram(self, tg_errors = None):
		self.removePageSession('mgmtSessionId', self.mgmtGridSessions)
		self.removePageSession('homeSessionId', self.homeGridSessions)
		self.removePageSession('rawSessionId', self.rawGridSessions)
		self.removePageSession('ttSessionId', self.ttGridSessions)
		self.removePageSession('histSessionId', self.histGridSessions)
		return dict(flotr=histogramFlotr,
					histForm=self.histogramForm,
					paging_grid=paging_grid,
					data_url = '/hist_grid_data',
					property_url='/hist_grid_settings')

	@expose(template="psychefe.templates.queryRawFlows")
	@identity.require(identity.in_any_group("admin", "analyst"))
	def queryRawFlows(self, tg_errors = None):
		self.rawFlowsParams = {}
		self.removePageSession('mgmtSessionId', self.mgmtGridSessions)
		self.removePageSession('homeSessionId', self.homeGridSessions)
		self.removePageSession('rawSessionId', self.rawGridSessions)
		self.removePageSession('ttSessionId', self.ttGridSessions)
		self.removePageSession('histSessionId', self.histGridSessions)
		return dict(paging_grid=paging_grid,
					data_url = '/raw_grid_data',
					property_url='/raw_grid_settings',
					rawFlowsForm=self.rawFlowsForm)

	@expose(template="psychefe.templates.management")
	@identity.require(identity.in_any_group("admin", "analyst"))
	def management(self, tg_errors = None):
		self.removePageSession('mgmtSessionId', self.mgmtGridSessions)
		self.removePageSession('homeSessionId', self.homeGridSessions)
		self.removePageSession('rawSessionId', self.rawGridSessions)
		self.removePageSession('ttSessionId', self.ttGridSessions)
		self.removePageSession('histSessionId', self.histGridSessions)
		return dict(paging_grid=paging_grid,
					data_url = '/mgmt_server_grid_data',
					property_url='/mgmt_server_grid_settings',
					mgmtForm=self.mgmtForm)

	#----------------------------------------------
	# Graphing Function
	# Prepares the json to pass to Flotr
	# in: data (dict)
	# in: graphType ('histogram', 'topTalkers', 'home')
	# out: json data prepared for Flotr
	#----------------------------------------------
	def getGraph(self, data, graphType):
		jsonData = {}
		
		if graphType == 'topTalkers':
			if data['dataseries']:
				if request.simple_cookie.has_key('ttSessionId'):
					sessionId = request.simple_cookie['ttSessionId'].value
					if sessionId in self.ttSessParams:
						topTalkersParams = self.ttSessParams[sessionId]
				else:
					return "Missing top talkers cookie"

				numSeries = len(data['dataseries'])
				info = data['datainfo']
			        items=info.items()
                	        backitems=[ [v[1],v[0]] for v in items]
                	        backitems.sort(reverse=True)
                	        info=[backitems[i][1] for i in range(0,len(backitems))] 
				bar_color = dict()
				bar_color['-2'] = 'white' # -2 indicates no data found for an interval
				i = 0;
				for talker in info:
					bar_color[talker] = ttColors[i]	
					i = i + 1

				data = self.xToTimeStamp(data['dataseries'], topTalkersParams['start_time'], topTalkersParams['time_interval'])
				#TODO: convert storedProcedure.py to do this
				delta = (data[0][1][0] - data[0][0][0]) - 1
				x_label = "x: Time <br/> y: Bytes"
				x_label += '<br/>' + self.getTopTalkersLegend() 

				# create the options passed to flotr
				options = {}
				options['bars']  =  { 'show' : 'true', 'barWidth' : delta - (delta/10), 'fillOpacity': 1 }
				options['mouse'] =  { 'track' : 'true', \
					  	      'sensibility' : delta, \
						      'trackFormatter' : 'dateBWTrackFormatter', \
						      'trackDecimals' : 4}
				options['xaxis'] =  { 'noTicks' : '10', \
                                                      'tickDecimals' : '1' , \
						      'tickFormatter' : 'timestampDHMTickFormatter' }
				options['yaxis'] =  { 'tickFormatter' : 'bwTickFormatter' }

				# now create the series passed to flotr
				temp_list = list(data[0])
				for i in range(len(temp_list)):
					temp_list[i] = [temp_list[i][0],0,'-2'] #initialize the list

				series = list()
				series.append({'data' : temp_list, 'label': x_label})
				for i in range(numSeries):
					idx = 0
					for d in data[i]:
						#logging.debug("data[" + str(i) + "]: " +str(d))
						temp_d = list(temp_list)
						temp_d[idx] = d
						series.append({'data' : temp_d, 'color' : bar_color[d[2]]})
						idx = idx + 1
				
				jsonData = {
						'series' : series,
						'options': options
					   }
			else:
				jsonData = {
						'series' : [{ 'data' : [], 'label' : 'No Data Returned' }],
						'options' : { 'legend' : { 'position' : 'ne' } }
					   }

		elif graphType == 'histogram':
			if data:
				if request.simple_cookie.has_key('histSessionId'):
					sessionId = request.simple_cookie['histSessionId'].value
					if sessionId in self.histSessParams:
						histParams = self.histSessParams[sessionId]
				else:
					return "Missing histogram cookie"

				ticks = []
				x_label = "x: In Bytes / Out Bytes <br/> y: log(Session Count)"
				#x_label += '<br/>Graph represents data for <br/>' + self.getHistogramLegend() 
				minX, maxX = self.getMinMax(data, 'x', True)
				delta = (maxX - minX) / len(data)
				d1 = data
				jsonData = {
						'series' : [{ 'data' : d1, 'label' : x_label, 
                                                              'bars' : { 'show' : 'true', 'barWidth': delta / 2 } }],
						'options': { 'mouse' : { 'track' : 'true',
									 'sensibility' : delta,
						      			 'trackFormatter' : 'histoTrackFormatter',
									 'trackDecimals' : 2, },
							     'legend' : { 'position' : 'ne' }, 
							     'xaxis' : { 'noTicks' : '10',
									 'tickFormatter': 'ratioTickFormatter' } }
					   }
			else:
				jsonData = {
						'series' : [{ 'data' : [], 'label' : 'No Data Returned' }],
						'options' : { 'legend' : { 'position' : 'ne' } }
					   }

		elif graphType == 'home':
			if data:
				if request.simple_cookie.has_key('homeSessionId'):
					sessionId = request.simple_cookie['homeSessionId'].value
					if sessionId in self.homeSessParams:
						homeParams = self.homeSessParams[sessionId]
				else:
					return "Missing home page cookie"

				x_label = "x: Time <br/> y: Bytes/sec"
				d1 = self.xToTimeStampHome(data, homeParams['start_time'], homeParams['time_interval'])
				delta = (d1[1][0] - data[0][0]) - 1
				jsonData = { 'series' : [{ 'data' : d1, 'label' : x_label, }],
					     'options': {
							'lines' : { 'show' : 'true' },
							'mouse' : { 'track' : 'true',
								    'sensibility' : delta,
								    'trackDecimals' : 0},
							'legend' : { 'position' : 'ne' }, 
							'xaxis' : {  'noTicks' : '10',
								     'tickDecimals' : '1',
								     'tickFormatter' : 'timestampDHMTickFormatter' },
							'yaxis' : { 'tickFormatter' : 'bwTickFormatter' } }
					   }
			else:
				jsonData = { 'series' : [{ 'data' : [], 'label' : 'No Data Returned' }],
							   'options' : { 'legend' : { 'position' : 'ne' } }
					   }

		return jsonData

	
	#----------------------------------------------
	# Graph Display Utility Functions
	# These functions help massage the data returned from the stored procedure for display
	# on the canvas. At some point these might be integrated into the SP object
	# --xToTimeStamp
	# --xToTimeStampHome
	# --getMinMax
	# --getHistogramLegend
	#----------------------------------------------
	#TODO: change x coord into timestamp, for now they're seperated into intervals
	def xToTimeStamp(self, data, start_time, interval):
		if data and start_time and interval:
			if interval == '4 hours':
			    intervalMinutes = 5 
			    minToAdd = 5
			elif interval == '1 day':
			    intervalMinutes = 30 
			    minToAdd = 30
			else: # 2 days
			    intervalMinutes = 60 
			    minToAdd = 60

			dataIdx = 0
			while dataIdx < len(data):
				idx = 0
				minToAdd = 0
				while idx < len(data[dataIdx]):
					data[dataIdx][idx][0] = time.mktime((start_time + datetime.timedelta(minutes=minToAdd)).timetuple())
					minToAdd += intervalMinutes
					idx += 1
				dataIdx += 1
			return data
		else:
			return data

	def xToTimeStampHome(self, data, start_time, interval):
		if data and start_time and interval:
			if interval == 'hour':
			    intervalMinutes = 5 
			    minToAdd = 5
			elif interval == 'day':
			    intervalMinutes = 5 
			    minToAdd = 5
			else: # week
			    intervalMinutes = 60 
			    minToAdd = 60

			for d in data:
				d[0] = time.mktime((start_time + datetime.timedelta(minutes=minToAdd)).timetuple())
				minToAdd += intervalMinutes
			return data
		else:
			return data

	#returns the min/max value for either the x or y coordinate in a coordinate pair list
	#default is to check the y coord, override by setting coord to 'x'
	def getMinMax(self, list, coord='y', includeZero=False):
		coords = {}
		minVal = 0
		maxVal = 0

		#put it into a dict
		for item in list:
			coords[item[0]] = item[1]

		if coord.upper() == 'X':
			coords = sorted(coords.keys())
		else:
			coords = sorted(coords.values())
		
		maxVal = coords[len(list) - 1]
		if includeZero:
			minVal = coords[0]
		else:
			for item in coords:
				if item > 0:
					minVal = item
					break;

		return minVal, maxVal

	def getTopTalkersLegend(self):
		retStr = "Bar Interval: "
		
		if request.simple_cookie.has_key('ttSessionId'):
			sessionId = request.simple_cookie['ttSessionId'].value
			if not (sessionId in self.ttSessParams):
				return ""
			else:
				ttParams = self.ttSessParams[sessionId]

		time_interval = ttParams['time_interval']
		if time_interval == '4 hours':
		  retStr += '5 minutes'
		elif time_interval == '1 day':
		  retStr += '30 minutes'
		else:
		  retStr += '1 hour'

		return retStr

	def getHistogramLegend(self):
		includedPorts = ""
		includedProts = ""
		excludedPorts = ""
		excludedProts = ""
		empty = False
		retStr = ""
		
		if request.simple_cookie.has_key('histSessionId'):
			sessionId = request.simple_cookie['histSessionId'].value
			if not (sessionId in self.histSessParams):
				return ""
			else:
				histogramParams = self.histSessParams[sessionId]

		#start with the included
		if histogramParams['inc_port_ranges'] != 'null':
			includedPorts += ', '.join(histogramParams['inc_port_ranges'].strip('array[[]]').replace(',','-').split(']-[')) + ", "
		if histogramParams['inc_prot_ranges'] != 'null':
			includedProts += ', '.join(histogramParams['inc_prot_ranges'].strip('array[[]]').replace(',','-').split(']-[')) + ", "
		if histogramParams['inc_ports'] != 'null':
			includedPorts += histogramParams['inc_ports'].strip('array[]').replace(',', ', ') + ", "
		if histogramParams['inc_prots'] != 'null':
			includedProts += histogramParams['inc_prots'].strip('array[]').replace(',', ', ') + ", "

		#now the excluded
		if histogramParams['exc_port_ranges'] != 'null':
			excludedPorts += ', '.join(histogramParams['exc_port_ranges'].strip('array[[]]').replace(',','-').split(']-[')) + ", "
		if histogramParams['exc_prot_ranges'] != 'null':
			excludedProts += ', '.join(histogramParams['exc_prot_ranges'].strip('array[[]]').replace(',','-').split(']-[')) + ", "
		if histogramParams['exc_ports'] != 'null':
			excludedPorts += histogramParams['exc_ports'].strip('array[]').replace(',', ', ') + ", "
		if histogramParams['exc_prots'] != 'null':
			excludedProts += histogramParams['exc_prots'].strip('array[]').replace(',', ', ') + ", "

		#oops, forgot about this one, no filters at all
		if len(includedPorts) + len(includedProts) + len(excludedPorts) + len(excludedProts) == 0:
			retStr += "All ports and protocols <br/>"	
		#ok, we only have port filters
		elif len(includedProts) == 0 and len(excludedProts) == 0:
			if len(includedPorts) > 0:
				retStr += "Port(s): " + includedPorts.rstrip(', ') + " <br/>"
				if len(excludedPorts) > 0:
					retStr += "But not port(s): " + excludedPorts.rstrip(', ') + " <br/>"
			elif len(excludedPorts) > 0:
				retStr += "All port(s) except: " + excludedPorts.rstrip(', ') + " <br/>"
		#no port filters, just protocol filters
		elif len(includedPorts) == 0 and len(excludedPorts) == 0:
			if len(includedProts) > 0:
				retStr += "Protocol(s): " + includedProts.rstrip(', ') + " <br/>"
				if len(excludedProts) > 0:
					retStr += "But not protocols(s): " + excludedPorts.rstrip(', ') + " <br/>"
			elif len(excludedProts) > 0:
				retStr += "All protocol(s) except: " + excludedProts.rstrip(', ') + " <br/>"
		#we have both port and protocol filters
		else:
			if len(includedPorts) > 0:
				retStr += "Port(s): " + includedPorts.rstrip(', ') + " <br/>"
				if len(excludedPorts) > 0:
					retStr += "But not port(s): " + excludedPorts.rstrip(', ') + " <br/>"
			elif len(excludedPorts) > 0:
				retStr += "All port(s) except: " + excludedPorts.rstrip(', ') + " <br/>"

			if len(includedProts) > 0:
				retStr += "Protocol(s): " + includedProts.rstrip(', ') + " <br/>"
				if len(excludedProts) > 0:
					retStr += "But not protocols(s): " + excludedPorts.rstrip(', ') + " <br/>"
			elif len(excludedProts) > 0:
				retStr += "All protocol(s) except: " + excludedProts.rstrip(', ') + " <br/>"
			
		return retStr.strip('<br/>')

	#----------------------------------------------
	# End Graph Display Utility Functions
	#----------------------------------------------

	#----------------------------------------------
	# Our Remote Form Error display handler
	# TODO: Look into moving elsewhere
	#----------------------------------------------
	@expose(template="psychefe.templates.topTalkers")
	def displayError(self, tg_errors = None):
		self.removePageSession('homeSessionId', self.homeGridSessions)
		self.removePageSession('rawSessionId', self.rawGridSessions)
		self.removePageSession('ttSessionId', self.ttGridSessions)
		self.removePageSession('histSessionId', self.histGridSessions)
		if type(tg_errors) == str:
			retStr = tg_errors
		else:
			retStr = ""
			areas = { 'targetNetworkArea'	:	'Target Network',
				  'timeArea'		:	'Time Area',
				  'graphOptionsArea'	:	'Graph Options Area' }
			fields = { 'network'		:	'Network of Interest',
				   'mgmt_network'	:       'Network to Add or Remove',
				   'src_network'	:       'Src Network of Interest',
				   'dst_network'	:       'Dst Network of Interest',
				   'start_time'		:	'Start Time',
				   'end_time'		:	'End Time',
				   'min_ratio'		:	'Min Ratio',
				   'max_ratio'		:	'Max Ratio'}
			for area in tg_errors.keys():
				retStr += areas[area] + ":<br/>"
				for field in tg_errors[area].keys():
					retStr += "Error in field '" + fields[field] + "': " + tg_errors[area][field] + "<br/>"
			retStr += '<style type="text/css">#barGraph img {display:none;}</style>'
		return retStr

	#----------------------------------------------
	# Our RemoteForm Ajax Callback URLs
	# --SubmitHistogramForm
	# --SubmitRawFlowsForm
	# --SubmitTopTalkersForm
	#----------------------------------------------
	@expose()
	@error_handler(displayError)
	@validate(form=histogramForm)
	def SubmitHistogramForm(self, tg_random, **kw):
		retStr = ""	
	
		if (len(kw) > 0):
			#move those nested dictionaries out
			keys = kw.keys()
			for key in keys:
				kw.update(kw[key])
				kw.pop(key)

			#ports and protocols
			if kw['portIncList']:
				kw['inc_port_ranges'], kw['inc_ports'] = self.buildArray(kw['portIncList'],'array[',']')
			else:
				kw['inc_port_ranges'] = 'null'
				kw['inc_ports'] = 'null'
			if kw['portExcList']:
				kw['exc_port_ranges'], kw['exc_ports'] = self.buildArray(kw['portExcList'],'array[',']')
			else:
				kw['exc_port_ranges'] = 'null'
				kw['exc_ports'] = 'null'
			if kw['protIncList']:
				kw['inc_prot_ranges'], kw['inc_prots'] = self.buildArray(kw['protIncList'],'array[',']')
			else:
				kw['inc_prot_ranges'] = 'null'
				kw['inc_prots'] = 'null'
			if kw['protExcList']:
				kw['exc_prot_ranges'], kw['exc_prots'] = self.buildArray(kw['protExcList'],'array[',']')
			else:
				kw['exc_prot_ranges'] = 'null'
				kw['exc_prots'] = 'null'

			#router and network data
			if kw['router']:
				kw['router'] = "'" + kw['router'] + "'"
			else:
				kw['router'] = "'none'"
			if kw['network']:
				kw['network'] = "'" + kw['network'] + "'"
			else:
				kw['network'] = "'none'"

			#clean up the Nones
			for key in kw.keys():
				if kw[key] == None:
					kw[key] = 'null'

			#make sure our datetime objs have second/microseconds
			kw['start_time'] = kw['start_time'].replace(second=01, microsecond=000001)
			kw['end_time'] = kw['end_time'].replace(second=01, microsecond=000001)

			#TODO: Create Validator
			if (kw['min_ratio'] == 'null' or kw['max_ratio'] == 'null' or type(kw['min_ratio']) != type(.1) or type(kw['max_ratio'] != type(.1))) and (not(kw['min_ratio'] > 0 and kw['max_ratio'] > 0 and (kw['min_ratio'] < kw['max_ratio']))):
				kw['min_ratio'] = .5
				kw['max_ratio'] = 2.0

			#self.numColumns = (kw['end_time'] - kw['start_time']).seconds / interval
			kw['num_buckets'] = '100'

			#set the default column and sort order for the grid
			kw['column'] = "'ratio'"
			kw['order'] = "'ASC'"

			#update the cookies
			self.removePageSession('histSessionId', self.histGridSessions, False)

			#get cursor for ajax grid
			resultsCursor = storedProcedures.CursorControl(kw, 'get_sessions')
			csvCursor = storedProcedures.CursorControl(kw, 'get_sessions')
			self.updatePageSession('histSessionId', tg_random)

			#let the server know about the session for the grid
			self.histGridSessions[tg_random] = resultsCursor
			self.histGridSessions[tg_random + "1"] = csvCursor

			#set parameters
			self.histSessParams[tg_random] = kw

		return retStr

	@expose()
	@error_handler(displayError)
	@validate(form=rawFlowsForm)
	def SubmitRawFlowsForm(self, tg_random, **kw):
		rawFlowsParams = {}
		retStr = ""	
	
		if (len(kw) > 0):
			#move those nested dictionaries out
			keys = kw.keys()
			for key in keys:
				kw.update(kw[key])
				kw.pop(key)

			#ports and protocols
			if kw['portIncList']:
				kw['inc_port_ranges'], kw['inc_ports'] = self.buildArray(kw['portIncList'],'array[',']')
			else:
				kw['inc_port_ranges'] = 'null'
				kw['inc_ports'] = 'null'
			if kw['portExcList']:
				kw['exc_port_ranges'], kw['exc_ports'] = self.buildArray(kw['portExcList'],'array[',']')
			else:
				kw['exc_port_ranges'] = 'null'
				kw['exc_ports'] = 'null'
			if kw['protIncList']:
				kw['inc_prot_ranges'], kw['inc_prots'] = self.buildArray(kw['protIncList'],'array[',']')
			else:
				kw['inc_prot_ranges'] = 'null'
				kw['inc_prots'] = 'null'
			if kw['protExcList']:
				kw['exc_prot_ranges'], kw['exc_prots'] = self.buildArray(kw['protExcList'],'array[',']')
			else:
				kw['exc_prot_ranges'] = 'null'
				kw['exc_prots'] = 'null'

			#router and network data
			if kw['router']:
				kw['router'] = "'" + kw['router'] + "'"
			else:
				kw['router'] = "'none'"
			if kw['src_network']:
				kw['src_network'] = "'" + kw['src_network'] + "'"
			else:
				kw['src_network'] = "'none'"
			if kw['dst_network']:
				kw['dst_network'] = "'" + kw['dst_network'] + "'"
			else:
				kw['dst_network'] = "'none'"

			#clean up the Nones
			for key in kw.keys():
				if kw[key] == None:
					kw[key] = 'null'

			#make sure our datetime objs have second/microseconds
			kw['start_time'] = kw['start_time'].replace(second=01, microsecond=000001)
			kw['end_time'] = kw['end_time'].replace(second=01, microsecond=000001)

			#set the default column and sort order for the grid
			kw['column'] = "'id'"
			kw['order'] = "'ASC'"

			#get cursor
			self.removePageSession('rawSessionId', self.rawGridSessions, False)
			resultsCursor = storedProcedures.CursorControl(kw, 'get_raw_flows')
			csvCursor = storedProcedures.CursorControl(kw, 'get_raw_flows')
			self.updatePageSession('rawSessionId', tg_random)

			#let the server know about the session
			self.rawGridSessions[tg_random] = resultsCursor
			self.rawGridSessions[tg_random + "1"] = csvCursor

			#set parameters
			self.rawSessParams[tg_random] = kw

		return retStr

	@expose()
	@error_handler(displayError)
	@validate(form=topTalkersForm)
	def SubmitTopTalkersForm(self, tg_random, **kw):
		retStr = ""	
	
		if (len(kw) > 0):
			#move those nested dictionaries out
			keys = kw.keys()
			for key in keys:
				kw.update(kw[key])
				kw.pop(key)

			#validate properly later
			if kw['direction'] == 1:
				kw['direction'] = "'src'"
			elif kw['direction'] == 2:
				kw['direction'] = "'dst'"

			#ports and protocols
			if kw['portIncList']:
				kw['inc_port_ranges'], kw['inc_ports'] = self.buildArray(kw['portIncList'],'array[',']')
			else:
				kw['inc_port_ranges'] = 'null'
				kw['inc_ports'] = 'null'
			if kw['portExcList']:
				kw['exc_port_ranges'], kw['exc_ports'] = self.buildArray(kw['portExcList'],'array[',']')
			else:
				kw['exc_port_ranges'] = 'null'
				kw['exc_ports'] = 'null'
			if kw['protIncList']:
				kw['inc_prot_ranges'], kw['inc_prots'] = self.buildArray(kw['protIncList'],'array[',']')
			else:
				kw['inc_prot_ranges'] = 'null'
				kw['inc_prots'] = 'null'
			if kw['protExcList']:
				kw['exc_prot_ranges'], kw['exc_prots'] = self.buildArray(kw['protExcList'],'array[',']')
			else:
				kw['exc_prot_ranges'] = 'null'
				kw['exc_prots'] = 'null'
			
			#routers & network mini validation
			if kw['router']:
				kw['router'] = "'" + kw['router'] + "'"
			else:
				kw['router'] = "'none'"
			if kw['network']:
				kw['network'] = "'" + kw['network'] + "'"
			else:
				kw['network'] = "'none'"

			#talker type
			if kw['talker_type'] == 1:
				kw['talker_type'] = "'port'"
			else:
				kw['talker_type'] = "'host'"

			#clean up the Nones, test
			for key in kw.keys():
				if kw[key] == None:
					kw[key] = 'null'

			#make sure our datetime objs have second/microseconds
			kw['start_time'] = kw['start_time'].replace(second=01, microsecond=000001)
			kw['end_time'] = kw['end_time'].replace(second=01, microsecond=000001)

			four_hours = datetime.timedelta(hours=4)
			one_day = datetime.timedelta(days=1)
			time_delta = kw['end_time'] - kw['start_time']

			if time_delta <= four_hours:
			    kw['time_interval'] = '4 hours'			 
			elif time_delta <= one_day:
			    kw['time_interval'] = '1 day'			 
			else:
			    kw['time_interval'] = '2 days'			 

			#set the default column and sort order for the grid
			kw['column'] = "'interval_start'"
			kw['order'] = "'DESC'"

			#get set parameters
			self.ttSessParams[tg_random] = kw

			#update the cookies
			self.removePageSession('ttSessionId', self.ttGridSessions, False)
			resultsCursor = storedProcedures.CursorControl(kw, 'get_traffic')
			csvCursor = storedProcedures.CursorControl(kw, 'get_traffic')
			self.updatePageSession('ttSessionId', tg_random)

			#let the server know about the session for the grid
			self.ttGridSessions[tg_random] = resultsCursor
			self.ttGridSessions[tg_random + "1"] = csvCursor

			#set parameters
			self.ttSessParams[tg_random] = kw

		return retStr

	@expose()
	@error_handler(displayError)
	@validate(form=homeForm)
	#def SubmitHomeForm(self, tg_random=str(int(random.random() * 10000000000000)), **kw):
	def SubmitHomeForm(self, tg_random, **kw):
		retStr = ""	

		if (len(kw) > 0):
		    #move those nested dictionaries out
		    keys = kw.keys()
		    for key in keys:
			    kw.update(kw[key])
			    kw.pop(key)
    
		    #set start and end times based on user selection
		    end_time = datetime.datetime.now()
		    if kw['time_interval'] == 1: # hour
			kw['time_interval'] = 'hour'
		        start_time = end_time - datetime.timedelta(hours=1)
		    elif kw['time_interval'] == 2: # day
			kw['time_interval'] = 'day'
		        start_time = end_time - datetime.timedelta(days=1)
		    else: # week
			kw['time_interval'] = 'week'
		        start_time = end_time - datetime.timedelta(weeks=1)

		    #shorten value that will be passed to SP
		    if kw['internal_net'] == "all internal networks":
			kw['internal_net'] = "all"
		    
		    #set direction values based on user selection
		    if kw['direction'] == 1: # source
			kw['direction'] = 'src'
		    else: # destination
			kw['direction'] = 'dst'
		    
		    #make sure our datetime objs have second/microseconds
		    kw['start_time'] = start_time.replace(second=01, microsecond=000001)
		    kw['end_time'] = end_time.replace(second=01, microsecond=000001)

		    #update the cookies
		    self.removePageSession('homeSessionId', self.homeGridSessions, False)
		    kw['order'] = "'ASC'"
		    kw['column'] = "'port'"
		    self.homeSessParams[tg_random] = kw
		    serverResultsCursor = storedProcedures.CursorControl(kw, 'get_internal_servers')
		    kw['order'] = "'DESC'"
		    kw['column'] = "'count'"
		    self.homeSessParams[tg_random + "1"] = kw
		    sessionResultsCursor = storedProcedures.CursorControl(kw, 'get_session_counts')
		    self.updatePageSession('homeSessionId', tg_random)
    
		    #let the server know about the session for the grids
		    self.homeGridSessions[tg_random] = serverResultsCursor
		    self.homeGridSessions[tg_random + "1"] = sessionResultsCursor
		
		return retStr

	@expose()
	@error_handler(displayError)
	@validate(form=mgmtForm)
	def SubmitMgmtForm(self, tg_random, **kw):
		retStr = ""	

		if (len(kw) > 0):
		    #move those nested dictionaries out
		    keys = kw.keys()
		    for key in keys:
			    kw.update(kw[key])
			    kw.pop(key)
    
		    if kw['mgmt_network']:
			kw['mgmt_network'] = "'" + kw['mgmt_network'] + "'"
		    else:
			kw['mgmt_network'] = "'none'"

		    if kw['action'] == 2: # add network
			self.dispatcher.dispatch('add_internal_net', kw)
		    elif kw['action'] == 3: # remove network
			self.dispatcher.dispatch('remove_internal_net', kw)

		    #update the cookies
		    self.removePageSession('mgmtSessionId', self.mgmtGridSessions, False)
		    self.mgmtSessParams[tg_random] = kw
		    kw['column'] = "'network'"
		    kw['order'] = "'ASC'"
		    serverResultsCursor = storedProcedures.CursorControl(kw, 'get_internal_nets')
		    self.updatePageSession('mgmtSessionId', tg_random)
    
		    #let the server know about the session for the grids
		    self.mgmtGridSessions[tg_random] = serverResultsCursor
		
		return retStr

	#----------------------------------------------
	# End Form Submit Urls
	#----------------------------------------------

	#utility function to remove a page session,
	#by default removes corresponding cookie, if you want to simply
	#update the cookie, pass in False and call updatePageSession
	#after calling this function.
	#Important: Always attempts to remove the session from dictToRem
	def removePageSession(self, name, dictToRem, removeCookie=True):
		if not removeCookie:
			if request.simple_cookie.has_key(name):
				if unicode(request.simple_cookie[name].value) in dictToRem:
					dictToRem[unicode(request.simple_cookie[name].value)].endSess()
					del dictToRem[unicode(request.simple_cookie[name].value)]
					csvFile = working_dir + str(unicode(request.simple_cookie[name].value)) + '.csv'
					if os.path.exists(csvFile):
						os.remove(csvFile)
		else:
			#get rid of our old cookie
			if request.simple_cookie.has_key(name):
				if unicode(request.simple_cookie[name].value) in dictToRem:
					dictToRem[unicode(request.simple_cookie[name].value)].endSess()
					del dictToRem[unicode(request.simple_cookie[name].value)]
				response.simple_cookie[name] = 0	
				response.simple_cookie[name]['expires'] = 0
				csvFile = working_dir + str(unicode(request.simple_cookie[name].value)) + '.csv'
				if os.path.exists(csvFile):
					os.remove(csvFile)
	
	#updates the cookie 'name' with the value 'newValue'
	def updatePageSession(self, name, newValue):
		response.simple_cookie[name] = newValue	
		response.simple_cookie[name]['path'] = '/'

	#TODO move to validator later
	#meaning, very little error checking done, get your data right
	#in data: "80,443,135-139,127-129"
	#out data: ranges: "[135,139],[127,129]"
	#		   singles: "80,443"
	#Note: if singles or ranges is empty by the end of the function will return 'null'
	#if supplied, preStr & postStr are appended to return variables
	def buildArray(self, inData, preStr=None, postStr=None):
		ranges = ""
		singles = ""
		rangePreStr = '['
		rangePostStr = ']'
		rangeSep = ','

		if preStr != None:
			singles = preStr
			ranges = preStr

		values = inData.split(',')

		#unique the list, only works in python 2.4+
		values = list(set(values))	

		for value in values:
			#get rid of the white space	
			value = value.strip()
			tmp = value.split('-')
			#did anything split
			if len(tmp) > 1:
				try:
					#check for validity, if there was more after tmp[1] we ignore, what are
					#they trying to do...
					if int(tmp[0]) > 0 and int(tmp[0]) < 65535 and int(tmp[1]) > int(tmp[0])\
					   and int(tmp[1]) < 65536:
						ranges += rangePreStr + tmp[0] + rangeSep + tmp[1] + rangePostStr + ',' 
				except:
					#silently fail
					continue
			else:
				try:
					if int(value) > 0 and int(value) < 65536:
						singles += value + ','
				except:
					#silently fail
					continue
		
		singles = singles.rstrip(',')
		ranges = ranges.rstrip(',')
		if postStr != None:
			singles += postStr
			ranges += postStr

		if singles == "":
			singles = 'null'
		elif singles == (preStr + postStr):
			singles = 'null'
		if ranges == "":
			ranges = 'null'
		elif ranges == (preStr + postStr):
			ranges = 'null'

		return str(ranges), str(singles)

	#----------------------------------------------
	# Helper functions to download the CSV files
	#----------------------------------------------
	@expose()
	def getFlowsCSV(self, **kw):
		if request.simple_cookie.has_key('rawSessionId'):
			sessionId = request.simple_cookie['rawSessionId'].value

			csvFile = working_dir + str(sessionId) + ".csv"
			if os.path.exists(csvFile):
				return serveFile(csvFile, "application/x-download", "attachment")

		return '<p align="center"><br/><b>Error - CSV file does not exist for this session.</b><br/><br/><a href="/queryRawFlows">Return to Raw Flows Page</a></p>'

	@expose()
	def getTopTalkersCSV(self, **kw):
		if request.simple_cookie.has_key('ttSessionId'):
			sessionId = request.simple_cookie['ttSessionId'].value

			csvFile = working_dir + str(sessionId) + ".csv"
			if os.path.exists(csvFile):
				return serveFile(csvFile, "application/x-download", "attachment")

		return '<p align="center"><br/><b>Error - CSV file does not exist for this session.</b><br/><br/><a href="/topTalkers">Return to Top Talkers Page</a></p>'

	@expose()
	def getHistCSV(self, **kw):
		if request.simple_cookie.has_key('histSessionId'):
			sessionId = request.simple_cookie['histSessionId'].value

			csvFile = working_dir + str(sessionId) + ".csv"
			if os.path.exists(csvFile):
				return serveFile(csvFile, "application/x-download", "attachment")

		return '<p align="center"><br/><b>Error - CSV file does not exist for this session.</b><br/><br/><a href="/histogram">Return to Histogram Page</a></p>'

	#----------------------------------------------
	# Our Histogram Sessions Display Ajaxgrid callback
	#----------------------------------------------
	@tg.expose("json")
	@tg.validate(validators={'start': validators.Int(),
							 'limit': validators.Int()})
	#@error_handler(displayError)
	def hist_grid_data(self, start=0, limit=25, sort=None, dir=None, **kw):
		total = 0
		items = []
		sessionId = ""
		columns = [
				'session_start',
				'last_activity',
				'router',
				'src_addr',
				'dst_addr',
				'src_port',
				'dst_port',
				'prot',
				'in_bytes',
				'out_bytes',
				'ratio'
			  ]

		#did the user submit the form properly and start a session?
		if request.simple_cookie.has_key('histSessionId'):
			sessionId = request.simple_cookie['histSessionId'].value
			kw = self.histSessParams[sessionId]

			if sessionId in self.histGridSessions:
				if sort != None:
				    sort = "'" + sort + "'"
				    dir  = "'" + dir  + "'"
				    if kw['column'] != sort or kw['order'] != dir:
					  kw['column'] = sort
					  kw['order'] = dir
					  self.histSessParams[sessionId] = kw
					  resultsCursor = storedProcedures.CursorControl(kw, 'get_sessions')
					  csvCursor = storedProcedures.CursorControl(kw, 'get_sessions')
					  self.histGridSessions[sessionId] = resultsCursor
					  self.histGridSessions[sessionId + "1"] = csvCursor

				dbCursor = self.histGridSessions[sessionId]
				total = str(int(dbCursor.count()))
				rows = dbCursor.getResults(start, start+limit)

				for row in rows[0:limit]: #possibly inherit from list class
					items.append({
						'session_start' : row[0],
						'last_activity' : row[1],
						'router' : row[2],
						'src_addr' : row[3],
						'dst_addr' : row[4],
						'src_port' : row[5],
						'dst_port' : row[6],
						'prot' : row[7],
						'in_bytes' : row[8],
						'out_bytes' : row[9],
						'ratio' : row[10]
					})

				# write the entire result set out to CSV
				csvCursor = self.histGridSessions[sessionId + "1"]
				total = int(csvCursor.count())
				rows = csvCursor.getResults(0, total)
				writer = csv.writer(open(str(sessionId) + ".csv", "wb"))
				writer.writerow(['session_start', 'last_activity', 'router', 'src_addr', 'dst_addr', 
						 'src_port', 'dst_port', 'prot', 'in_bytes', 'out_bytes', 'ratio'])
				for row in rows:
					writer.writerow([row[0], row[1], row[2], row[3], row[4], 
							 row[5], row[6], row[7], row[8], row[9], row[10]])
				
			else:
				raise redirect('/histogram')

		#return "Rows = " + dbCursor
		return dict(totalCount=total, items=items)

	#----------------------------------------------
	# Our Raw Flows Display Ajaxgrid callback
	#----------------------------------------------
	@tg.expose("json")
	@tg.validate(validators={'start': validators.Int(), 'limit': validators.Int()})
	def raw_grid_data(self, start=0, limit=25, sort=None, dir=None, **kw):
		total = 0
		items = []
		sessionId = ""
		columns = [
				'id',
				'router',
				'src_addr',
				'dst_addr',
				'start_time',
				'end_time',
				'tcp_flags',
				'prot',
				'tos',
				'input_if',
				'output_if',
				'src_as',
				'dst_as',
				'src_port',
				'dst_port',
				'src_mask',
				'dst_mask',
				'next_hop',
				'pkts',
				'bytes'
			  ]

		#did the user submit the form properly and start a session?
		if request.simple_cookie.has_key('rawSessionId'):
			sessionId = request.simple_cookie['rawSessionId'].value
			kw = self.rawSessParams[sessionId]

			if sessionId in self.rawGridSessions:
				if sort != None:
				    sort = "'" + sort + "'"
				    dir  = "'" + dir  + "'"
				    if kw['column'] != sort or kw['order'] != dir:
					  kw['column'] = sort
					  kw['order'] = dir
					  self.rawSessParams[sessionId] = kw
					  resultsCursor = storedProcedures.CursorControl(kw, 'get_raw_flows')
					  csvCursor = storedProcedures.CursorControl(kw, 'get_raw_flows')
					  self.rawGridSessions[sessionId] = resultsCursor
					  self.rawGridSessions[sessionId + "1"] = csvCursor

				dbCursor = self.rawGridSessions[sessionId]
				total = str(int(dbCursor.count()))
				rows = dbCursor.getResults(start, start+limit)

				for row in rows[0:limit]: #possibly inherit from list class
					items.append({
						'router' : row[1],
						'src_addr' : row[2],
						'dst_addr' : row[3],
						'start_time' : row[4],
						'end_time' : row[5],
						'tcp_flags' : row[6],
						'prot' : row[7],
#						'tos' : row[8],
#						'input_if' : row[9],
#						'output_if' : row[10],
#						'src_as' : row[11],
#						'dst_as' : row[12],
						'src_port' : row[13],
						'dst_port' : row[14],
#						'src_mask' : row[15],
#						'dst_mask' : row[16],
#						'next_hop' : row[17],
						'pkts' : row[18],
						'bytes' : row[19]
					})

				# write the entire result set out to CSV
				csvCursor = self.rawGridSessions[sessionId + "1"]
				total = int(csvCursor.count())
				rows = csvCursor.getResults(0, total)
				writer = csv.writer(open(str(sessionId) + ".csv", "wb"))
				writer.writerow(['router', 'src_addr', 'dst_addr', 'start_time', 'end_time', 
						 'tcp_flags', 'prot', 'src_port', 'dst_port', 'pkts', 'bytes'])
				for row in rows:
					writer.writerow([row[1], row[2], row[3], row[4], row[5], 
							 row[6], row[7], row[13], row[14], row[18], row[19]])
				
			else:
				raise redirect('/queryRawFlows')

		return dict(totalCount=total,
					items=items)

	#----------------------------------------------
	# Our Home Page Ajaxgrid callbacks
	#----------------------------------------------
	@tg.expose("json")
	@tg.validate(validators={'start': validators.Int(), 'limit': validators.Int()})
	def home_server_grid_data(self, start=0, limit=10, sort=None, dir=None, **kw):
		total = 0
		items = []
		sessionId = ""
		columns = [
				'router',
				'addr',
				'port'
			  ]

		#did the user submit the form properly and start a session?
		if request.simple_cookie.has_key('homeSessionId'):
			sessionId = request.simple_cookie['homeSessionId'].value
			kw = self.homeSessParams[sessionId]

			if sessionId in self.homeGridSessions:
				if sort != None:
				    sort = "'" + sort + "'"
				    dir  = "'" + dir  + "'"
				    if kw['column'] != sort or kw['order'] != dir:
					  kw['column'] = sort
					  kw['order'] = dir
					  self.homeSessParams[sessionId] = kw
					  resultsCursor = storedProcedures.CursorControl(kw, 'get_internal_servers')
					  self.homeGridSessions[sessionId] = resultsCursor

				dbCursor = self.homeGridSessions[sessionId]
				total = str(int(dbCursor.count()))
				rows = dbCursor.getResults(start, start+limit)

				for row in rows[0:limit]: #possibly inherit from list class
					items.append({
						'router' : row[0],
						'addr' : row[1],
						'port' : row[2]
					})
			else:
				raise redirect('/index')

		return dict(totalCount=total, items=items)

	@tg.expose("json")
	@tg.validate(validators={'start': validators.Int(), 'limit': validators.Int()})
	def home_session_grid_data(self, start=0, limit=10, sort=None, dir=None, **kw):
		total = 0
		items = []
		sessionId = ""
		columns = [
				'protocol',
				'count'
			  ]

		#did the user submit the form properly and start a session?
		if request.simple_cookie.has_key('homeSessionId'):
			sessionId = request.simple_cookie['homeSessionId'].value + "1"
			kw = self.homeSessParams[sessionId]

			if sessionId in self.homeGridSessions:
				if sort != None:
				    sort = "'" + sort + "'"
				    dir  = "'" + dir  + "'"
				    if kw['column'] != sort or kw['order'] != dir:
					  kw['column'] = sort
					  kw['order'] = dir
					  self.homeSessParams[sessionId] = kw
					  resultsCursor = storedProcedures.CursorControl(kw, 'get_session_counts')
					  self.homeGridSessions[sessionId] = resultsCursor

				dbCursor = self.homeGridSessions[sessionId]
				total = str(int(dbCursor.count()))
				rows = dbCursor.getResults(start, start+limit)

				for row in rows[0:limit]: #possibly inherit from list class
					items.append({
						'protocol' : row[0],
						'count' : row[1]
					})
			else:
				raise redirect('/index')

		return dict(totalCount=total, items=items)

	@tg.expose("json")
	@tg.validate(validators={'start': validators.Int(), 'limit': validators.Int()})
	def mgmt_server_grid_data(self, start=0, limit=10, sort=None, dir=None, **kw):
		total = 0
		items = []
		sessionId = ""
		columns = [
				'network'
			  ]

		#did the user submit the form properly and start a session?
		if request.simple_cookie.has_key('mgmtSessionId'):
			sessionId = request.simple_cookie['mgmtSessionId'].value
			kw = self.mgmtSessParams[sessionId]

			if sessionId in self.mgmtGridSessions:
				if sort != None:
				    sort = "'" + sort + "'"
				    dir  = "'" + dir  + "'"
				    if kw['column'] != sort or kw['order'] != dir:
					  kw['order'] = dir
					  self.mgmtSessParams[sessionId] = kw
					  resultsCursor = storedProcedures.CursorControl(kw, 'get_internal_nets')
					  self.mgmtGridSessions[sessionId] = resultsCursor

				dbCursor = self.mgmtGridSessions[sessionId]
				total = str(int(dbCursor.count()))
				rows = dbCursor.getResults(start, start+limit)

				for row in rows[0:limit]: #possibly inherit from list class
					items.append({
						'network' : row[0]
					})
			else:
				raise redirect('/management')

		return dict(totalCount=total, items=items)

	#----------------------------------------------
	# Our Top Talkers Traffic Display Ajaxgrid callback
	#----------------------------------------------
	@tg.expose("json")
	@tg.validate(validators={'start': validators.Int(), 'limit': validators.Int()})
	#@error_handler(displayError)
	def tt_grid_data(self, start=0, limit=25, sort=None, dir=None, **kw):
		total = 0
		items = []
		sessionId = ""
		columns = [
				'interval_start',
				'router',
				'addr',
				'prot',
				'port',
				'pkts',
				'bytes'
			  ]

		#did the user submit the form properly and start a session?
		if request.simple_cookie.has_key('ttSessionId'):
			sessionId = request.simple_cookie['ttSessionId'].value
			kw = self.ttSessParams[sessionId]

			if sessionId in self.ttGridSessions:
				if sort != None: 
				    sort = "'" + sort + "'"
				    dir  = "'" + dir  + "'"
				    if kw['column'] != sort or kw['order'] != dir:
					  kw['column'] = sort
					  kw['order'] = dir
					  self.ttSessParams[sessionId] = kw
					  resultsCursor = storedProcedures.CursorControl(kw, 'get_traffic')
					  csvCursor = storedProcedures.CursorControl(kw, 'get_traffic')
					  self.ttGridSessions[sessionId] = resultsCursor
					  self.ttGridSessions[sessionId + "1"] = csvCursor

				dbCursor = self.ttGridSessions[sessionId]
				total = str(int(dbCursor.count()))
				rows = dbCursor.getResults(start, start+limit)

				for row in rows[0:limit]: #possibly inherit from list class
				    items.append({
						'interval_start' : row[0],
						'router' : row[1],
						'addr' : row[2],
						'prot' : row[3],
						'port' : row[4],
						'pkts' : row[5],
						'bytes' : row[6]
					       })

				# write the entire result set out to CSV
				csvCursor = self.ttGridSessions[sessionId + "1"]
				total = int(csvCursor.count())
				rows = csvCursor.getResults(0, total)
				writer = csv.writer(open(str(sessionId) + ".csv", "wb"))
				writer.writerow(['interval_start', 'router', 'addr', 'prot', 'port', 'pkts', 'bytes'])
				for row in rows:
					writer.writerow([row[0], row[1], row[2], row[3], row[4], row[5], row[6]])

			else:
				raise redirect('/topTalkers')

		return dict(totalCount=total,
					items=items)

	def buildTTGrid(self, datainfo, talker_type):
		rows = []
		headers = []
		if len(datainfo) > 0:
			items=datainfo.items()
                	backitems=[ [v[1],v[0]] for v in items]
                	backitems.sort(reverse=True)
                	sortedlist=[backitems[i][1] for i in range(0,len(backitems))] 

			if talker_type == "'port'":
				headers.append('Port')
				other_code = '-1'
			else:
				headers.append('Host')
				other_code = '0.0.0.0'
			headers.append('KBytes')
			headers.append('Color')
			other_color = 'none'
			other_bw = 0
			[rows.append([]) for i in range(len(sortedlist))]
			i = 0
			j = 0
			for iidx in sortedlist:
				if iidx == other_code:
					other_color = ttColors[i]
					other_bw = datainfo[iidx] / 1000;
				else:
					rows[j].append(iidx)
					rows[j].append(datainfo[iidx] / 1000)
					rows[j].append(ttColors[i])
					j = j + 1
				i = i + 1
			rows[len(rows) - 1].append("All Other")
			rows[len(rows) - 1].append(other_bw)
			rows[len(rows) - 1].append(other_color)
		return headers, rows

	#----------------------------------------------
	# Our TopTalkers Ajaxgrid callback
	# TODO: Change page session mapping to cursor instead of params
	#	Create method for cursor control to fetch all results
	#	Create stacked bar graph class to hold methods use to 
	#	massage data into format for graph
	#----------------------------------------------
	@expose()
	def getInfo(self, **kw):
		return '<p align="center"><br/>Test</p>'

	@expose()
	@expose(format="json")
	def getColumnInfo(self, **kw):
		ttRows = []
		ttHeaders = []
		if request.simple_cookie.has_key('ttSessionId'):
			sessionId = request.simple_cookie['ttSessionId'].value

			if sessionId in self.ttSessParams:
				topTalkersParams = self.ttSessParams[sessionId]
				if (len(topTalkersParams) > 0):
					topTalkersData = self.dispatcher.dispatch('get_top_talkers', topTalkersParams)
					if topTalkersData != None:
						ttHeaders, ttRows = self.buildTTGrid(topTalkersData['datainfo'], topTalkersParams['talker_type'])
		return dict(headers=ttHeaders,rows=ttRows)

	#----------------------------------------------
	# Our TurboFlotr Ajax Callback URLs
	# --UpdateTopTalkers
	# --UpdateHistogram
	#----------------------------------------------
	@expose(format="json")
	def UpdateTopTalkers(self, **kw):
		series = {}
		topTalkersData = {}

		if request.simple_cookie.has_key('ttSessionId'):
			sessionId = request.simple_cookie['ttSessionId'].value

			if sessionId in self.ttSessParams:
				topTalkersParams = self.ttSessParams[sessionId]
				if (len(topTalkersParams) > 0):
					topTalkersData = self.dispatcher.dispatch('get_top_talkers', topTalkersParams)

				series = self.getGraph(topTalkersData, 'topTalkers')

		return dict(series=series)

	@expose(format="json")
	def UpdateHistogram(self, **kw):
		series = {}
		histogramData = {}
		sessionId = ""

		#did the user submit the form properly and start a session?
		if request.simple_cookie.has_key('histSessionId'):
			sessionId = request.simple_cookie['histSessionId'].value

			if sessionId in self.histSessParams:
				histogramParams = self.histSessParams[sessionId]
				if (len(histogramParams) > 0):
					histogramData = self.dispatcher.dispatch('bw_ratio', histogramParams)

				series = self.getGraph(histogramData, 'histogram')

		return dict(series=series)

	@expose(format="json")
	def UpdateHomeGraph(self, **kw):
		series = {}
		homeData = {}
		sessionId = ""
		
		#did the user submit the form properly and start a session?
		if request.simple_cookie.has_key('homeSessionId'):
			sessionId = request.simple_cookie['homeSessionId'].value

			if sessionId in self.homeSessParams:
				homeParams = self.homeSessParams[sessionId]
				if (len(homeParams) > 0):
					homeData = self.dispatcher.dispatch('bw_by_interval', homeParams)

				series = self.getGraph(homeData, 'home')

		return dict(series=series)

