import random ; random.seed()
import simplejson
import pkg_resources

from turbogears.widgets import CSSLink, JSLink, Widget
from turbogears.widgets import register_static_directory

__all__ = ["TurboFlotr"]

js_dir = pkg_resources.resource_filename("psychefe.widgets.turboflotr", "static")
register_static_directory("turboflotr", js_dir)

		#<div id="${id}" style="width:${width};height:${height};margin:2em 0 ${marginBottom} 20%;"></div>
class TurboFlotr(Widget):
	"""
		A TurboGears Flotr Widget. This widget loads its data from the submission of a
		form on the page.
	"""
	template = """

		<div xmlns:py="http://purl.org/kid/ns#">
		<div id="${id}" style="width:${width};height:${height};margin:auto;"></div>
		<script type="text/javascript">
		var typeGraph = null;
		${helperFunc}
		var f = null;
		Event.observe(document, 'dom:loaded', function() {
			document.observe('ajax:completed', function() {
					new Ajax.Request('${refresh_url}', {
								 method:'get',
							 	 onSuccess: function(transport) {

						/**
						* Parse (eval) the JSON from the server.
						*/
						var data = transport.responseText.evalJSON();

						/* More to follow */
						if (typeGraph == 'timeseries') {
							data.series.options.xaxis.tickFormatter = timestampDHMTickFormatter;
							data.series.options.yaxis.tickFormatter = bwTickFormatter;
							data.series.options.mouse.trackFormatter = dateBWTrackFormatter;
						}
						else {
							data.series.options.xaxis.tickFormatter = ratioTickFormatter;
							data.series.options.mouse.trackFormatter = histoTrackFormatter;
						}

						/**
						 * Draw the graph using the JSON data. Of course the
						 * options are optional.
						 */
						f = Flotr.draw($('${id}'), data.series.series, data.series.options);
					}
				});
			});
		});
		</script>
	  </div>
	"""

	params = ["series", "options", "height", "width", "marginBottom", "id", "label", "form_id", "refresh_url", "helperFunc"]
	params_doc = {
			"series"		: "An array of data series",
			"options"		: "Plot options",
			"height"		: "The height of the graph",
			"width"			: "The width of the graph",
			"marginBottom"	: "CSS Bottom Margin Value, default 0",
			"label"			: "Label for the graph",
			"id"			: "An optional ID for the graph",
			"form_id"		: "The form id that will submit the data to be graphed",
			"refresh_url"	: "URL of the update function",
			"helperFunc"	: "Any additonal javascript functions to add to the widget"
	}
	css = [CSSLink('turboflotr', 'turboflotr.css')]
	javascript = [#JSLink('turboflotr', 'lib/excanvas.js'),
				  JSLink("turboflotr", "lib/prototype-1.6.0.2.js"),
				  JSLink("turboflotr", "flotr.js")]

	def __init__(self, refresh_url, form_id, series={}, options={}, height="300px", width="600px",
				 id=None, label='', marginBottom = '0', helperFunc=None):
		super(TurboFlotr, self).__init__()
		if id:
			self.id = id
		else:
			self.id = "turboflotr" + str(int(random.random() * 1000))
		self.series = simplejson.dumps(series)
		self.options = simplejson.dumps(options)
		self.height = height
		self.width = width
		self.label = label
		self.refresh_url = refresh_url
		self.form_id = form_id
		self.marginBottom = marginBottom
		self.helperFunc = helperFunc

